"""Comprehensive tests for the rate limiting module."""

import pytest
import time
import threading
from unittest.mock import Mock, patch

from truthound.ratelimit import (
    # Core types
    RateLimitStrategy,
    RateLimitScope,
    RateLimitAction,
    RateLimitResult,
    RateLimitConfig,
    QuotaConfig,
    RateLimitExceeded,
    QuotaExceeded,
    # Key extractors
    GlobalKeyExtractor,
    AttributeKeyExtractor,
    CompositeKeyExtractor,
    CallableKeyExtractor,
    # Algorithms
    TokenBucketAlgorithm,
    SlidingWindowAlgorithm,
    FixedWindowAlgorithm,
    LeakyBucketAlgorithm,
    create_algorithm,
    # Storage
    MemoryStorage,
    create_storage,
    # Policies
    PolicyPriority,
    DefaultPolicy,
    TierBasedPolicy,
    EndpointPolicy,
    IPBasedPolicy,
    CompositePolicy,
    ConditionalPolicy,
    QuotaManager,
    PolicyRegistry,
    DynamicPolicyConfig,
    # Main limiter
    RateLimiter,
    RateLimiterRegistry,
    rate_limit,
    RateLimitContext,
    # Middleware
    RateLimitMetrics,
    MetricsMiddleware,
    RetryHandler,
)


# =============================================================================
# Test Rate Limit Result
# =============================================================================


class TestRateLimitResult:
    """Tests for RateLimitResult."""

    def test_result_creation(self):
        """Test creating a rate limit result."""
        result = RateLimitResult(
            allowed=True,
            remaining=5,
            limit=10,
            reset_at=time.time() + 60,
        )
        assert result.allowed is True
        assert result.remaining == 5
        assert result.limit == 10

    def test_utilization(self):
        """Test utilization calculation."""
        result = RateLimitResult(
            allowed=True,
            remaining=3,
            limit=10,
            reset_at=time.time(),
        )
        assert result.utilization == 0.7

    def test_utilization_zero_limit(self):
        """Test utilization with zero limit."""
        result = RateLimitResult(
            allowed=False,
            remaining=0,
            limit=0,
            reset_at=time.time(),
        )
        assert result.utilization == 0.0

    def test_to_headers(self):
        """Test header generation."""
        now = time.time()
        result = RateLimitResult(
            allowed=True,
            remaining=5,
            limit=10,
            reset_at=now + 60,
        )
        headers = result.to_headers()
        assert "RateLimit-Limit" in headers
        assert headers["RateLimit-Limit"] == "10"
        assert headers["RateLimit-Remaining"] == "5"

    def test_to_headers_retry_after(self):
        """Test headers include Retry-After when not allowed."""
        result = RateLimitResult(
            allowed=False,
            remaining=0,
            limit=10,
            reset_at=time.time() + 60,
            retry_after=30,
        )
        headers = result.to_headers()
        assert "Retry-After" in headers
        assert headers["Retry-After"] == "30"


# =============================================================================
# Test Rate Limit Config
# =============================================================================


class TestRateLimitConfig:
    """Tests for RateLimitConfig."""

    def test_default_config(self):
        """Test default configuration."""
        config = RateLimitConfig()
        assert config.requests_per_second == 10.0
        assert config.burst_size == 10

    def test_custom_config(self):
        """Test custom configuration."""
        config = RateLimitConfig(
            requests_per_second=100,
            burst_size=200,
            strategy=RateLimitStrategy.SLIDING_WINDOW,
        )
        assert config.requests_per_second == 100
        assert config.burst_size == 200
        assert config.strategy == RateLimitStrategy.SLIDING_WINDOW

    def test_refill_rate(self):
        """Test refill rate property."""
        config = RateLimitConfig(requests_per_second=50)
        assert config.refill_rate == 50

    def test_from_dict(self):
        """Test creating config from dictionary."""
        data = {
            "requests_per_second": 20,
            "strategy": "sliding_window",
        }
        config = RateLimitConfig.from_dict(data)
        assert config.requests_per_second == 20
        assert config.strategy == RateLimitStrategy.SLIDING_WINDOW

    def test_invalid_config(self):
        """Test validation of invalid config."""
        with pytest.raises(ValueError):
            RateLimitConfig(requests_per_second=-1)


# =============================================================================
# Test Key Extractors
# =============================================================================


class TestKeyExtractors:
    """Tests for key extractors."""

    def test_global_key_extractor(self):
        """Test global key extractor."""
        extractor = GlobalKeyExtractor("my_key")
        assert extractor.extract({}) == "my_key"
        assert extractor.extract(None) == "my_key"

    def test_attribute_key_extractor_dict(self):
        """Test attribute extractor with dict."""
        extractor = AttributeKeyExtractor("user_id", prefix="user:")
        context = {"user_id": "123"}
        assert extractor.extract(context) == "user:123"

    def test_attribute_key_extractor_object(self):
        """Test attribute extractor with object."""
        extractor = AttributeKeyExtractor("user_id")

        class Context:
            user_id = "456"

        assert extractor.extract(Context()) == "456"

    def test_attribute_key_extractor_default(self):
        """Test attribute extractor with default."""
        extractor = AttributeKeyExtractor("missing", default="default")
        assert extractor.extract({}) == "default"

    def test_composite_key_extractor(self):
        """Test composite key extractor."""
        extractor = CompositeKeyExtractor([
            AttributeKeyExtractor("endpoint"),
            AttributeKeyExtractor("user_id"),
        ], separator=":")
        context = {"endpoint": "/api", "user_id": "123"}
        assert extractor.extract(context) == "/api:123"

    def test_callable_key_extractor(self):
        """Test callable key extractor."""
        extractor = CallableKeyExtractor(lambda ctx: f"ip:{ctx.get('ip')}")
        assert extractor.extract({"ip": "1.2.3.4"}) == "ip:1.2.3.4"


# =============================================================================
# Test Token Bucket Algorithm
# =============================================================================


class TestTokenBucketAlgorithm:
    """Tests for Token Bucket algorithm."""

    def test_basic_acquire(self):
        """Test basic token acquisition."""
        config = RateLimitConfig(requests_per_second=10, burst_size=10)
        bucket = TokenBucketAlgorithm(config)

        result = bucket.acquire("test")
        assert result.allowed is True
        assert result.remaining == 9

    def test_burst_limit(self):
        """Test burst limit enforcement."""
        config = RateLimitConfig(requests_per_second=10, burst_size=5)
        bucket = TokenBucketAlgorithm(config)

        # Consume all burst capacity
        for _ in range(5):
            result = bucket.acquire("test")
            assert result.allowed is True

        # Should be rejected
        result = bucket.acquire("test")
        assert result.allowed is False
        assert result.retry_after > 0

    def test_token_refill(self):
        """Test token refill over time."""
        config = RateLimitConfig(requests_per_second=100, burst_size=5)
        bucket = TokenBucketAlgorithm(config)

        # Consume all tokens
        for _ in range(5):
            bucket.acquire("test")

        # Wait for refill
        time.sleep(0.1)  # Should refill ~10 tokens

        result = bucket.acquire("test")
        assert result.allowed is True

    def test_peek_no_consume(self):
        """Test peek doesn't consume tokens."""
        config = RateLimitConfig(requests_per_second=10, burst_size=5)
        bucket = TokenBucketAlgorithm(config)

        result1 = bucket.peek("test")
        result2 = bucket.peek("test")

        assert result1.remaining == result2.remaining

    def test_reset(self):
        """Test bucket reset."""
        config = RateLimitConfig(requests_per_second=10, burst_size=5)
        bucket = TokenBucketAlgorithm(config)

        # Consume some tokens
        bucket.acquire("test")
        bucket.acquire("test")

        # Reset
        bucket.reset("test")

        # Should be full again
        result = bucket.peek("test")
        assert result.remaining == 5

    def test_multiple_keys(self):
        """Test separate limits per key."""
        config = RateLimitConfig(requests_per_second=10, burst_size=2)
        bucket = TokenBucketAlgorithm(config)

        # Exhaust key1
        bucket.acquire("key1")
        bucket.acquire("key1")
        result1 = bucket.acquire("key1")

        # key2 should still have tokens
        result2 = bucket.acquire("key2")

        assert result1.allowed is False
        assert result2.allowed is True


# =============================================================================
# Test Sliding Window Algorithm
# =============================================================================


class TestSlidingWindowAlgorithm:
    """Tests for Sliding Window algorithm."""

    def test_basic_acquire(self):
        """Test basic acquisition."""
        config = RateLimitConfig(
            requests_per_second=10,
            window_size_seconds=1.0,
            sub_window_count=10,
        )
        window = SlidingWindowAlgorithm(config)

        result = window.acquire("test")
        assert result.allowed is True

    def test_window_limit(self):
        """Test window limit enforcement."""
        config = RateLimitConfig(
            requests_per_second=5,
            window_size_seconds=1.0,
        )
        window = SlidingWindowAlgorithm(config)

        # Consume all capacity
        for _ in range(5):
            result = window.acquire("test")
            assert result.allowed is True

        # Should be rejected
        result = window.acquire("test")
        assert result.allowed is False

    def test_window_slide(self):
        """Test window sliding."""
        config = RateLimitConfig(
            requests_per_second=100,  # 10 per 0.1 second
            window_size_seconds=0.2,
            sub_window_count=2,
        )
        window = SlidingWindowAlgorithm(config)

        # Consume some requests
        for _ in range(10):
            window.acquire("test")

        # Wait for sub-window to expire
        time.sleep(0.15)

        # Should have some capacity again
        result = window.acquire("test")
        assert result.allowed is True


# =============================================================================
# Test Fixed Window Algorithm
# =============================================================================


class TestFixedWindowAlgorithm:
    """Tests for Fixed Window algorithm."""

    def test_basic_acquire(self):
        """Test basic acquisition."""
        config = RateLimitConfig(
            requests_per_second=10,
            window_size_seconds=1.0,
        )
        window = FixedWindowAlgorithm(config)

        result = window.acquire("test")
        assert result.allowed is True

    def test_window_reset(self):
        """Test window reset at boundary."""
        config = RateLimitConfig(
            requests_per_second=100,  # 10 per 0.1 second
            window_size_seconds=0.1,
        )
        window = FixedWindowAlgorithm(config)

        # Consume all capacity
        for _ in range(10):
            window.acquire("test")

        # Wait for new window
        time.sleep(0.15)

        # Should be allowed in new window
        result = window.acquire("test")
        assert result.allowed is True


# =============================================================================
# Test Leaky Bucket Algorithm
# =============================================================================


class TestLeakyBucketAlgorithm:
    """Tests for Leaky Bucket algorithm."""

    def test_basic_acquire(self):
        """Test basic acquisition."""
        config = RateLimitConfig(
            requests_per_second=10,
            burst_size=10,
        )
        bucket = LeakyBucketAlgorithm(config)

        result = bucket.acquire("test")
        assert result.allowed is True

    def test_queue_capacity(self):
        """Test queue capacity limit."""
        config = RateLimitConfig(
            requests_per_second=1,  # Slow leak rate
            burst_size=5,
        )
        bucket = LeakyBucketAlgorithm(config)

        # Fill queue
        for _ in range(5):
            result = bucket.acquire("test")
            assert result.allowed is True

        # Should be rejected (queue full)
        result = bucket.acquire("test")
        assert result.allowed is False


# =============================================================================
# Test Algorithm Factory
# =============================================================================


class TestAlgorithmFactory:
    """Tests for algorithm factory."""

    def test_create_token_bucket(self):
        """Test creating token bucket."""
        config = RateLimitConfig(strategy=RateLimitStrategy.TOKEN_BUCKET)
        algorithm = create_algorithm(config)
        assert isinstance(algorithm, TokenBucketAlgorithm)

    def test_create_sliding_window(self):
        """Test creating sliding window."""
        config = RateLimitConfig(strategy=RateLimitStrategy.SLIDING_WINDOW)
        algorithm = create_algorithm(config)
        assert isinstance(algorithm, SlidingWindowAlgorithm)

    def test_create_fixed_window(self):
        """Test creating fixed window."""
        config = RateLimitConfig(strategy=RateLimitStrategy.FIXED_WINDOW)
        algorithm = create_algorithm(config)
        assert isinstance(algorithm, FixedWindowAlgorithm)

    def test_create_leaky_bucket(self):
        """Test creating leaky bucket."""
        config = RateLimitConfig(strategy=RateLimitStrategy.LEAKY_BUCKET)
        algorithm = create_algorithm(config)
        assert isinstance(algorithm, LeakyBucketAlgorithm)


# =============================================================================
# Test Memory Storage
# =============================================================================


class TestMemoryStorage:
    """Tests for memory storage."""

    def test_get_set(self):
        """Test basic get/set."""
        storage = MemoryStorage()
        storage.set("key", "value")
        assert storage.get("key") == "value"

    def test_get_missing(self):
        """Test getting missing key."""
        storage = MemoryStorage()
        assert storage.get("missing") is None

    def test_delete(self):
        """Test deletion."""
        storage = MemoryStorage()
        storage.set("key", "value")
        assert storage.delete("key") is True
        assert storage.get("key") is None
        assert storage.delete("key") is False

    def test_increment(self):
        """Test atomic increment."""
        storage = MemoryStorage()
        assert storage.increment("counter") == 1
        assert storage.increment("counter") == 2
        assert storage.increment("counter", 5) == 7

    def test_ttl_expiration(self):
        """Test TTL expiration."""
        storage = MemoryStorage()
        storage.set("key", "value", ttl=0.1)
        assert storage.get("key") == "value"
        time.sleep(0.15)
        assert storage.get("key") is None

    def test_get_with_lock(self):
        """Test get with lock."""
        storage = MemoryStorage()
        storage.set("key", "value")

        value, lock = storage.get_with_lock("key")
        assert value == "value"
        assert lock is not None

        # Release lock
        storage.set_with_lock("key", "new_value", lock)
        assert storage.get("key") == "new_value"

    def test_clear(self):
        """Test clearing storage."""
        storage = MemoryStorage()
        storage.set("key1", "value1")
        storage.set("key2", "value2")
        storage.clear()
        assert storage.size() == 0


# =============================================================================
# Test Policies
# =============================================================================


class TestPolicies:
    """Tests for rate limit policies."""

    def test_default_policy(self):
        """Test default policy."""
        config = RateLimitConfig(requests_per_second=10)
        policy = DefaultPolicy(config)

        match = policy.matches({})
        assert match.matched is True
        assert match.config == config

    def test_tier_based_policy(self):
        """Test tier-based policy."""
        policy = TierBasedPolicy(
            tier_configs={
                "free": RateLimitConfig(requests_per_second=1),
                "pro": RateLimitConfig(requests_per_second=10),
            },
            tier_extractor=lambda ctx: ctx.get("tier", "free"),
        )

        # Free tier
        match = policy.matches({"tier": "free"})
        assert match.matched is True
        assert match.config.requests_per_second == 1

        # Pro tier
        match = policy.matches({"tier": "pro"})
        assert match.matched is True
        assert match.config.requests_per_second == 10

    def test_endpoint_policy(self):
        """Test endpoint policy."""
        policy = EndpointPolicy(
            endpoint_configs={
                "/api/search": RateLimitConfig(requests_per_second=5),
                "/api/export": RateLimitConfig(requests_per_second=1),
            },
            endpoint_extractor=lambda ctx: ctx.get("path"),
        )

        match = policy.matches({"path": "/api/search"})
        assert match.matched is True
        assert match.config.requests_per_second == 5

    def test_ip_based_policy(self):
        """Test IP-based policy."""
        policy = IPBasedPolicy(
            ip_extractor=lambda ctx: ctx.get("ip"),
            default_config=RateLimitConfig(requests_per_second=10),
            allowlist=["192.168.1.1"],
            allowlist_config=RateLimitConfig(requests_per_second=100),
        )

        # Allowlisted IP
        match = policy.matches({"ip": "192.168.1.1"})
        assert match.config.requests_per_second == 100

        # Normal IP
        match = policy.matches({"ip": "10.0.0.1"})
        assert match.config.requests_per_second == 10

    def test_composite_policy(self):
        """Test composite policy."""
        tier_policy = TierBasedPolicy(
            tier_configs={"vip": RateLimitConfig(requests_per_second=100)},
            tier_extractor=lambda ctx: ctx.get("tier"),
        )
        default_policy = DefaultPolicy(RateLimitConfig(requests_per_second=10))

        policy = CompositePolicy([tier_policy, default_policy])

        # VIP matches tier policy
        match = policy.matches({"tier": "vip"})
        assert match.config.requests_per_second == 100

        # Non-VIP falls through to default
        match = policy.matches({"tier": "regular"})
        assert match.config.requests_per_second == 10

    def test_conditional_policy(self):
        """Test conditional policy."""
        policy = ConditionalPolicy(
            condition=lambda ctx: ctx.get("is_admin", False),
            config=RateLimitConfig(requests_per_second=1000),
        )

        # Admin
        match = policy.matches({"is_admin": True})
        assert match.matched is True

        # Non-admin
        match = policy.matches({"is_admin": False})
        assert match.matched is False


# =============================================================================
# Test Quota Manager
# =============================================================================


class TestQuotaManager:
    """Tests for quota manager."""

    def test_register_quota(self):
        """Test registering a quota."""
        manager = QuotaManager()
        quota = QuotaConfig(
            name="api_calls",
            limit=1000,
            period_seconds=86400,
        )
        manager.register_quota(quota)

        assert manager.get_quota("api_calls") is not None

    def test_consume_quota(self):
        """Test consuming quota."""
        manager = QuotaManager()
        manager.register_quota(QuotaConfig(
            name="api_calls",
            limit=10,
            period_seconds=86400,
        ))

        usage = manager.consume("user:123", "api_calls", 5)
        assert usage.used == 5
        assert usage.remaining == 5

    def test_quota_exceeded(self):
        """Test quota exceeded exception."""
        manager = QuotaManager()
        manager.register_quota(QuotaConfig(
            name="api_calls",
            limit=5,
            period_seconds=86400,
        ))

        # Consume all quota
        manager.consume("user:123", "api_calls", 5)

        # Should raise
        with pytest.raises(QuotaExceeded):
            manager.consume("user:123", "api_calls", 1)

    def test_quota_check(self):
        """Test checking quota without consuming."""
        manager = QuotaManager()
        manager.register_quota(QuotaConfig(
            name="api_calls",
            limit=10,
            period_seconds=86400,
        ))

        result = manager.check("user:123", "api_calls", 5)
        assert result.allowed is True

        # Should still be allowed (didn't consume)
        result = manager.check("user:123", "api_calls", 5)
        assert result.allowed is True


# =============================================================================
# Test Policy Registry
# =============================================================================


class TestPolicyRegistry:
    """Tests for policy registry."""

    def test_register_resolve(self):
        """Test registering and resolving policies."""
        registry = PolicyRegistry()
        registry.register(DefaultPolicy(RateLimitConfig(requests_per_second=10)))

        match = registry.resolve({})
        assert match.matched is True

    def test_unregister(self):
        """Test unregistering policies."""
        registry = PolicyRegistry()
        policy = DefaultPolicy(RateLimitConfig(), name="test")
        registry.register(policy)

        assert registry.unregister("test") is True
        assert registry.unregister("test") is False


# =============================================================================
# Test Rate Limiter
# =============================================================================


class TestRateLimiter:
    """Tests for main RateLimiter class."""

    def test_basic_usage(self):
        """Test basic rate limiter usage."""
        limiter = RateLimiter(
            config=RateLimitConfig(requests_per_second=10, burst_size=5)
        )

        result = limiter.acquire("test")
        assert result.allowed is True

    def test_key_extraction(self):
        """Test key extraction from context."""
        limiter = RateLimiter(
            config=RateLimitConfig(requests_per_second=10),
            key_extractor=AttributeKeyExtractor("user_id"),
        )

        result = limiter.acquire(context={"user_id": "123"})
        assert result.allowed is True
        assert result.bucket_key == "123"

    def test_peek(self):
        """Test peeking without consuming."""
        limiter = RateLimiter(
            config=RateLimitConfig(requests_per_second=10, burst_size=5)
        )

        result1 = limiter.peek("test")
        result2 = limiter.peek("test")
        assert result1.remaining == result2.remaining

    def test_reset(self):
        """Test resetting limiter."""
        limiter = RateLimiter(
            config=RateLimitConfig(requests_per_second=10, burst_size=5)
        )

        limiter.acquire("test")
        limiter.acquire("test")
        limiter.reset("test")

        result = limiter.peek("test")
        assert result.remaining == 5


# =============================================================================
# Test Rate Limiter Registry
# =============================================================================


class TestRateLimiterRegistry:
    """Tests for rate limiter registry."""

    def setup_method(self):
        """Reset registry before each test."""
        RateLimiterRegistry.reset_instance()

    def test_register_get(self):
        """Test registering and getting limiters."""
        registry = RateLimiterRegistry()
        limiter = RateLimiter(name="test")
        registry.register("test", limiter)

        assert registry.get("test") is limiter

    def test_default_limiter(self):
        """Test default limiter."""
        registry = RateLimiterRegistry()
        limiter = RateLimiter(name="test")
        registry.register("test", limiter, set_default=True)

        assert registry.get() is limiter

    def test_get_or_create(self):
        """Test get or create."""
        registry = RateLimiterRegistry()

        limiter = registry.get_or_create("new")
        assert limiter is not None
        assert registry.get("new") is limiter


# =============================================================================
# Test Decorator
# =============================================================================


class TestRateLimitDecorator:
    """Tests for rate limit decorator."""

    def setup_method(self):
        """Reset registry before each test."""
        RateLimiterRegistry.reset_instance()

    def test_basic_decorator(self):
        """Test basic decorator usage."""
        limiter = RateLimiter(
            config=RateLimitConfig(requests_per_second=10, burst_size=5),
            name="test",
        )

        @rate_limit(key="test_func", limiter=limiter)
        def my_function():
            return "success"

        result = my_function()
        assert result == "success"

    def test_decorator_with_key_func(self):
        """Test decorator with key function."""
        limiter = RateLimiter(
            config=RateLimitConfig(requests_per_second=10, burst_size=5),
            name="test",
        )

        @rate_limit(
            key=lambda user_id: f"user:{user_id}",
            limiter=limiter,
        )
        def my_function(user_id):
            return f"success:{user_id}"

        result = my_function("123")
        assert result == "success:123"

    def test_decorator_exceeded(self):
        """Test decorator raises on exceeded."""
        limiter = RateLimiter(
            config=RateLimitConfig(requests_per_second=10, burst_size=1),
            name="test",
        )

        @rate_limit(key="test", limiter=limiter)
        def my_function():
            return "success"

        # First call succeeds
        my_function()

        # Second call should raise
        with pytest.raises(RateLimitExceeded):
            my_function()


# =============================================================================
# Test Context Manager
# =============================================================================


class TestRateLimitContext:
    """Tests for rate limit context manager."""

    def test_context_manager_allowed(self):
        """Test context manager when allowed."""
        limiter = RateLimiter(
            config=RateLimitConfig(requests_per_second=10, burst_size=5)
        )

        with RateLimitContext("test", limiter=limiter) as result:
            assert result.allowed is True

    def test_context_manager_exceeded(self):
        """Test context manager when exceeded."""
        limiter = RateLimiter(
            config=RateLimitConfig(requests_per_second=10, burst_size=1)
        )

        # Exhaust limit
        limiter.acquire("test")

        with pytest.raises(RateLimitExceeded):
            with RateLimitContext("test", limiter=limiter):
                pass

    def test_context_manager_no_raise(self):
        """Test context manager without raising."""
        limiter = RateLimiter(
            config=RateLimitConfig(requests_per_second=10, burst_size=1)
        )

        limiter.acquire("test")

        with RateLimitContext("test", limiter=limiter, raise_on_exceeded=False) as result:
            assert result.allowed is False


# =============================================================================
# Test Metrics
# =============================================================================


class TestRateLimitMetrics:
    """Tests for rate limit metrics."""

    def test_record_metrics(self):
        """Test recording metrics."""
        metrics = RateLimitMetrics()

        # Record allowed
        metrics.record(RateLimitResult(
            allowed=True, remaining=5, limit=10, reset_at=time.time()
        ))

        # Record rejected
        metrics.record(RateLimitResult(
            allowed=False, remaining=0, limit=10, reset_at=time.time()
        ))

        assert metrics.total_requests == 2
        assert metrics.allowed_requests == 1
        assert metrics.rejected_requests == 1
        assert metrics.rejection_rate == 0.5

    def test_metrics_middleware(self):
        """Test metrics middleware."""
        limiter = RateLimiter(
            config=RateLimitConfig(requests_per_second=10, burst_size=5)
        )
        middleware = MetricsMiddleware(limiter)

        middleware.acquire("test")
        middleware.acquire("test")

        assert middleware.metrics.total_requests == 2
        assert middleware.metrics.allowed_requests == 2


# =============================================================================
# Test Retry Handler
# =============================================================================


class TestRetryHandler:
    """Tests for retry handler."""

    def test_immediate_success(self):
        """Test immediate success without retry."""
        limiter = RateLimiter(
            config=RateLimitConfig(requests_per_second=10, burst_size=5)
        )
        handler = RetryHandler(limiter, max_retries=3)

        def my_func():
            return "success"

        result = handler.execute(my_func, key="test")
        assert result == "success"

    def test_retry_on_limit(self):
        """Test retry when rate limited."""
        limiter = RateLimiter(
            config=RateLimitConfig(requests_per_second=100, burst_size=1)  # Fast refill
        )
        handler = RetryHandler(
            limiter,
            max_retries=3,
            base_delay=0.05,
        )

        # Exhaust limit
        limiter.acquire("test")

        call_count = 0

        def my_func():
            nonlocal call_count
            call_count += 1
            return "success"

        result = handler.execute(my_func, key="test")
        assert result == "success"


# =============================================================================
# Test Thread Safety
# =============================================================================


class TestThreadSafety:
    """Tests for thread safety."""

    def test_concurrent_acquire(self):
        """Test concurrent token acquisition."""
        config = RateLimitConfig(requests_per_second=1000, burst_size=100)
        limiter = RateLimiter(config=config)

        results = []
        errors = []

        def worker():
            try:
                for _ in range(10):
                    result = limiter.acquire("test")
                    results.append(result)
            except Exception as e:
                errors.append(e)

        threads = [threading.Thread(target=worker) for _ in range(10)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        assert len(errors) == 0
        assert len(results) == 100

    def test_concurrent_storage(self):
        """Test concurrent storage access."""
        storage = MemoryStorage()

        def worker(worker_id):
            for i in range(100):
                storage.increment(f"counter:{worker_id}")

        threads = [threading.Thread(target=worker, args=(i,)) for i in range(10)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        # Each worker should have count of 100
        for i in range(10):
            state = storage.get(f"counter:{i}")
            assert state is not None


# =============================================================================
# Test Dynamic Policy Config
# =============================================================================


class TestDynamicPolicyConfig:
    """Tests for dynamic policy configuration."""

    def test_get_default_config(self):
        """Test getting default config."""
        config = DynamicPolicyConfig(
            default_config=RateLimitConfig(requests_per_second=10)
        )

        result = config.get_config("unknown_key")
        assert result.requests_per_second == 10

    def test_override_config(self):
        """Test override configuration."""
        config = DynamicPolicyConfig(
            default_config=RateLimitConfig(requests_per_second=10)
        )

        config.set_override("vip", RateLimitConfig(requests_per_second=100))

        result = config.get_config("vip")
        assert result.requests_per_second == 100

    def test_remove_override(self):
        """Test removing override."""
        config = DynamicPolicyConfig(
            default_config=RateLimitConfig(requests_per_second=10)
        )

        config.set_override("vip", RateLimitConfig(requests_per_second=100))
        config.remove_override("vip")

        result = config.get_config("vip")
        assert result.requests_per_second == 10


# =============================================================================
# Test Storage Factory
# =============================================================================


class TestStorageFactory:
    """Tests for storage factory."""

    def test_create_memory_storage(self):
        """Test creating memory storage."""
        storage = create_storage("memory")
        assert isinstance(storage, MemoryStorage)

    def test_create_unknown_storage(self):
        """Test creating unknown storage type."""
        with pytest.raises(ValueError):
            create_storage("unknown")



"""Tests for Microsoft Teams notification action.

This module contains comprehensive tests for the Teams notification
functionality including Adaptive Card building, templates, and HTTP handling.
"""

from __future__ import annotations

import json
from datetime import datetime
from typing import Any
from unittest.mock import MagicMock, Mock, patch

import pytest


# =============================================================================
# Test Fixtures
# =============================================================================


@pytest.fixture
def mock_checkpoint_result() -> MagicMock:
    """Create a mock CheckpointResult for testing."""
    from truthound.checkpoint.actions.base import ActionStatus

    # Create mock validation statistics
    mock_stats = MagicMock()
    mock_stats.total_issues = 5
    mock_stats.critical_issues = 1
    mock_stats.high_issues = 2
    mock_stats.medium_issues = 1
    mock_stats.low_issues = 1
    mock_stats.pass_rate = 0.85

    # Create mock validation result
    mock_validation = MagicMock()
    mock_validation.statistics = mock_stats
    mock_validation.results = [
        MagicMock(
            validator_name="NotNullValidator",
            column="email",
            message="Found 2 null values",
        ),
        MagicMock(
            validator_name="RangeValidator",
            column="age",
            message="Found 1 value out of range",
        ),
    ]

    # Create mock checkpoint result
    mock_result = MagicMock()
    mock_result.checkpoint_name = "test_checkpoint"
    mock_result.run_id = "run_abc123def456"
    mock_result.run_time = datetime(2025, 12, 27, 10, 30, 0)
    mock_result.duration_ms = 1234.56
    mock_result.data_asset = "customers.csv"
    mock_result.validation_result = mock_validation
    mock_result.status = MagicMock(value="failure")

    return mock_result


@pytest.fixture
def mock_success_result() -> MagicMock:
    """Create a mock successful CheckpointResult."""
    mock_stats = MagicMock()
    mock_stats.total_issues = 0
    mock_stats.critical_issues = 0
    mock_stats.high_issues = 0
    mock_stats.medium_issues = 0
    mock_stats.low_issues = 0
    mock_stats.pass_rate = 1.0

    mock_validation = MagicMock()
    mock_validation.statistics = mock_stats
    mock_validation.results = []

    mock_result = MagicMock()
    mock_result.checkpoint_name = "test_checkpoint"
    mock_result.run_id = "run_success123"
    mock_result.run_time = datetime(2025, 12, 27, 10, 30, 0)
    mock_result.duration_ms = 500.0
    mock_result.data_asset = "customers.csv"
    mock_result.validation_result = mock_validation
    mock_result.status = MagicMock(value="success")

    return mock_result


# =============================================================================
# Adaptive Card Builder Tests
# =============================================================================


class TestAdaptiveCardBuilder:
    """Tests for AdaptiveCardBuilder class."""

    def test_basic_card_build(self):
        """Test building a basic Adaptive Card."""
        from truthound.checkpoint.actions.teams_notify import AdaptiveCardBuilder

        builder = AdaptiveCardBuilder()
        builder.add_text_block("Hello World")

        card = builder.build()

        assert card["type"] == "AdaptiveCard"
        assert card["$schema"] == "http://adaptivecards.io/schemas/adaptive-card.json"
        assert card["version"] == "1.4"
        assert len(card["body"]) == 1
        assert card["body"][0]["type"] == "TextBlock"
        assert card["body"][0]["text"] == "Hello World"

    def test_version_setting(self):
        """Test setting card version."""
        from truthound.checkpoint.actions.teams_notify import (
            AdaptiveCardBuilder,
            AdaptiveCardVersion,
        )

        builder = AdaptiveCardBuilder(AdaptiveCardVersion.V1_5)
        card = builder.build()

        assert card["version"] == "1.5"

    def test_text_block_options(self):
        """Test text block with various options."""
        from truthound.checkpoint.actions.teams_notify import (
            AdaptiveCardBuilder,
            TextWeight,
            TextSize,
            TextColor,
        )

        builder = AdaptiveCardBuilder()
        builder.add_text_block(
            "Styled Text",
            weight=TextWeight.BOLDER,
            size=TextSize.LARGE,
            color=TextColor.ATTENTION,
            is_subtle=True,
            max_lines=2,
            horizontal_alignment="center",
        )

        card = builder.build()
        text_block = card["body"][0]

        assert text_block["text"] == "Styled Text"
        assert text_block["weight"] == "bolder"
        assert text_block["size"] == "large"
        assert text_block["color"] == "attention"
        assert text_block["isSubtle"] is True
        assert text_block["maxLines"] == 2
        assert text_block["horizontalAlignment"] == "center"

    def test_fact_set(self):
        """Test adding a fact set."""
        from truthound.checkpoint.actions.teams_notify import AdaptiveCardBuilder

        builder = AdaptiveCardBuilder()
        builder.add_fact_set([
            ("Name", "John Doe"),
            ("Status", "Active"),
            ("Role", "Admin"),
        ])

        card = builder.build()
        fact_set = card["body"][0]

        assert fact_set["type"] == "FactSet"
        assert len(fact_set["facts"]) == 3
        assert fact_set["facts"][0] == {"title": "Name", "value": "John Doe"}

    def test_container(self):
        """Test adding a container."""
        from truthound.checkpoint.actions.teams_notify import (
            AdaptiveCardBuilder,
            CardContainerStyle,
        )

        builder = AdaptiveCardBuilder()
        builder.add_container(
            [{"type": "TextBlock", "text": "Inside container"}],
            style=CardContainerStyle.EMPHASIS,
            bleed=True,
        )

        card = builder.build()
        container = card["body"][0]

        assert container["type"] == "Container"
        assert container["style"] == "emphasis"
        assert container["bleed"] is True

    def test_column_set(self):
        """Test adding a column set."""
        from truthound.checkpoint.actions.teams_notify import AdaptiveCardBuilder

        columns = [
            {
                "type": "Column",
                "width": "auto",
                "items": [{"type": "TextBlock", "text": "Col 1"}],
            },
            {
                "type": "Column",
                "width": "stretch",
                "items": [{"type": "TextBlock", "text": "Col 2"}],
            },
        ]

        builder = AdaptiveCardBuilder()
        builder.add_column_set(columns)

        card = builder.build()
        column_set = card["body"][0]

        assert column_set["type"] == "ColumnSet"
        assert len(column_set["columns"]) == 2

    def test_action_open_url(self):
        """Test adding an OpenUrl action."""
        from truthound.checkpoint.actions.teams_notify import AdaptiveCardBuilder

        builder = AdaptiveCardBuilder()
        builder.add_action_open_url(
            "View Details",
            "https://example.com",
            style="positive",
            tooltip="Click to view",
        )

        card = builder.build()

        assert "actions" in card
        assert len(card["actions"]) == 1

        action = card["actions"][0]
        assert action["type"] == "Action.OpenUrl"
        assert action["title"] == "View Details"
        assert action["url"] == "https://example.com"
        assert action["style"] == "positive"
        assert action["tooltip"] == "Click to view"

    def test_action_submit(self):
        """Test adding a Submit action."""
        from truthound.checkpoint.actions.teams_notify import AdaptiveCardBuilder

        builder = AdaptiveCardBuilder()
        builder.add_action_submit("Submit", data={"key": "value"})

        card = builder.build()
        action = card["actions"][0]

        assert action["type"] == "Action.Submit"
        assert action["data"] == {"key": "value"}

    def test_action_show_card(self):
        """Test adding a ShowCard action."""
        from truthound.checkpoint.actions.teams_notify import AdaptiveCardBuilder

        nested_card = {"type": "AdaptiveCard", "body": []}

        builder = AdaptiveCardBuilder()
        builder.add_action_show_card("Show More", nested_card)

        card = builder.build()
        action = card["actions"][0]

        assert action["type"] == "Action.ShowCard"
        assert action["card"] == nested_card

    def test_image(self):
        """Test adding an image."""
        from truthound.checkpoint.actions.teams_notify import AdaptiveCardBuilder

        builder = AdaptiveCardBuilder()
        builder.add_image(
            "https://example.com/image.png",
            alt_text="Example Image",
            size="medium",
            style="person",
        )

        card = builder.build()
        image = card["body"][0]

        assert image["type"] == "Image"
        assert image["url"] == "https://example.com/image.png"
        assert image["altText"] == "Example Image"
        assert image["size"] == "medium"
        assert image["style"] == "person"

    def test_full_width(self):
        """Test enabling full width."""
        from truthound.checkpoint.actions.teams_notify import AdaptiveCardBuilder

        builder = AdaptiveCardBuilder()
        builder.enable_full_width()

        card = builder.build()

        assert card["msteams"]["width"] == "Full"

    def test_fallback_text(self):
        """Test setting fallback text."""
        from truthound.checkpoint.actions.teams_notify import AdaptiveCardBuilder

        builder = AdaptiveCardBuilder()
        builder.set_fallback_text("Fallback message for old clients")

        card = builder.build()

        assert card["fallbackText"] == "Fallback message for old clients"

    def test_build_message_card(self):
        """Test building a complete message card."""
        from truthound.checkpoint.actions.teams_notify import AdaptiveCardBuilder

        builder = AdaptiveCardBuilder()
        builder.add_text_block("Test message")

        message = builder.build_message_card()

        assert message["type"] == "message"
        assert len(message["attachments"]) == 1
        assert message["attachments"][0]["contentType"] == "application/vnd.microsoft.card.adaptive"
        assert message["attachments"][0]["content"]["type"] == "AdaptiveCard"

    def test_fluent_api(self):
        """Test that builder supports fluent API."""
        from truthound.checkpoint.actions.teams_notify import AdaptiveCardBuilder

        card = (
            AdaptiveCardBuilder()
            .set_version("1.4")
            .add_text_block("Title", weight="bolder")
            .add_text_block("Subtitle", is_subtle=True)
            .add_fact_set([("Key", "Value")])
            .add_action_open_url("Click", "https://example.com")
            .build()
        )

        assert len(card["body"]) == 3
        assert len(card["actions"]) == 1


# =============================================================================
# Message Template Tests
# =============================================================================


class TestMessageTemplates:
    """Tests for message template system."""

    def test_get_template_default(self):
        """Test getting default template."""
        from truthound.checkpoint.actions.teams_notify import (
            get_template,
            DefaultTemplate,
            MessageTheme,
        )

        template = get_template(MessageTheme.DEFAULT)
        assert isinstance(template, DefaultTemplate)

        template = get_template("default")
        assert isinstance(template, DefaultTemplate)

    def test_get_template_minimal(self):
        """Test getting minimal template."""
        from truthound.checkpoint.actions.teams_notify import (
            get_template,
            MinimalTemplate,
            MessageTheme,
        )

        template = get_template(MessageTheme.MINIMAL)
        assert isinstance(template, MinimalTemplate)

    def test_get_template_detailed(self):
        """Test getting detailed template."""
        from truthound.checkpoint.actions.teams_notify import (
            get_template,
            DetailedTemplate,
            MessageTheme,
        )

        template = get_template(MessageTheme.DETAILED)
        assert isinstance(template, DetailedTemplate)

    def test_get_template_compact(self):
        """Test getting compact template."""
        from truthound.checkpoint.actions.teams_notify import (
            get_template,
            CompactTemplate,
            MessageTheme,
        )

        template = get_template(MessageTheme.COMPACT)
        assert isinstance(template, CompactTemplate)

    def test_get_template_invalid(self):
        """Test getting invalid template raises error."""
        from truthound.checkpoint.actions.teams_notify import get_template

        with pytest.raises(ValueError, match="Unknown theme"):
            get_template("nonexistent")

    def test_register_custom_template(self):
        """Test registering a custom template."""
        from truthound.checkpoint.actions.teams_notify import (
            register_template,
            get_template,
            MessageTemplate,
        )

        class CustomTemplate(MessageTemplate):
            def render(self, checkpoint_result, config):
                return {"type": "message", "custom": True}

        register_template("my_custom", CustomTemplate)
        template = get_template("my_custom")

        assert isinstance(template, CustomTemplate)

    def test_default_template_render(self, mock_checkpoint_result):
        """Test rendering default template."""
        from truthound.checkpoint.actions.teams_notify import (
            DefaultTemplate,
            TeamsConfig,
        )

        template = DefaultTemplate()
        config = TeamsConfig(
            webhook_url="https://test.webhook.office.com/test",
            include_details=True,
            include_actions=False,
        )

        message = template.render(mock_checkpoint_result, config)

        assert message["type"] == "message"
        assert len(message["attachments"]) == 1

        content = message["attachments"][0]["content"]
        assert content["type"] == "AdaptiveCard"
        assert len(content["body"]) > 0

    def test_minimal_template_render(self, mock_checkpoint_result):
        """Test rendering minimal template."""
        from truthound.checkpoint.actions.teams_notify import (
            MinimalTemplate,
            TeamsConfig,
        )

        template = MinimalTemplate()
        config = TeamsConfig(webhook_url="https://test.webhook.office.com/test")

        message = template.render(mock_checkpoint_result, config)

        assert message["type"] == "message"
        content = message["attachments"][0]["content"]
        # Minimal template should have fewer body elements
        assert len(content["body"]) <= 2

    def test_compact_template_render(self, mock_checkpoint_result):
        """Test rendering compact template."""
        from truthound.checkpoint.actions.teams_notify import (
            CompactTemplate,
            TeamsConfig,
        )

        template = CompactTemplate()
        config = TeamsConfig(webhook_url="https://test.webhook.office.com/test")

        message = template.render(mock_checkpoint_result, config)

        assert message["type"] == "message"
        content = message["attachments"][0]["content"]
        # Compact uses column set
        assert content["body"][0]["type"] == "ColumnSet"

    def test_template_with_mentions(self, mock_checkpoint_result):
        """Test template with @mentions on failure."""
        from truthound.checkpoint.actions.teams_notify import (
            DefaultTemplate,
            TeamsConfig,
        )

        template = DefaultTemplate()
        config = TeamsConfig(
            webhook_url="https://test.webhook.office.com/test",
            mention_on_failure=[
                {"id": "user@example.com", "name": "John Doe"},
                "another@example.com",
            ],
        )

        message = template.render(mock_checkpoint_result, config)
        content = message["attachments"][0]["content"]

        # Check that mentions entities were added
        assert "msteams" in content
        assert "entities" in content["msteams"]
        assert len(content["msteams"]["entities"]) == 2

    def test_template_with_action_buttons(self, mock_checkpoint_result):
        """Test template with action buttons."""
        from truthound.checkpoint.actions.teams_notify import (
            DefaultTemplate,
            TeamsConfig,
        )

        template = DefaultTemplate()
        config = TeamsConfig(
            webhook_url="https://test.webhook.office.com/test",
            include_actions=True,
            dashboard_url="https://dashboard.example.com/runs/{run_id}",
            details_url="https://details.example.com/{checkpoint}",
        )

        message = template.render(mock_checkpoint_result, config)
        content = message["attachments"][0]["content"]

        assert "actions" in content
        assert len(content["actions"]) == 2

        # Verify URL substitution
        assert "run_abc123def456" in content["actions"][0]["url"]


# =============================================================================
# TeamsConfig Tests
# =============================================================================


class TestTeamsConfig:
    """Tests for TeamsConfig dataclass."""

    def test_default_config(self):
        """Test default configuration values."""
        from truthound.checkpoint.actions.teams_notify import (
            TeamsConfig,
            TeamsConnectorType,
            MessageTheme,
            AdaptiveCardVersion,
            NotifyCondition,
        )

        config = TeamsConfig()

        assert config.webhook_url == ""
        assert config.connector_type == TeamsConnectorType.INCOMING_WEBHOOK
        assert config.theme == MessageTheme.DEFAULT
        assert config.card_version == AdaptiveCardVersion.V1_4
        assert config.full_width is True
        assert config.include_details is True
        assert config.include_actions is True
        assert config.mention_on_failure == []
        assert config.notify_on == NotifyCondition.FAILURE

    def test_config_with_string_enums(self):
        """Test config converts string enums properly."""
        from truthound.checkpoint.actions.teams_notify import (
            TeamsConfig,
            TeamsConnectorType,
            AdaptiveCardVersion,
        )

        config = TeamsConfig(
            webhook_url="https://test.webhook.office.com/test",
            connector_type="incoming_webhook",
            card_version="1.5",
        )

        assert config.connector_type == TeamsConnectorType.INCOMING_WEBHOOK
        assert config.card_version == AdaptiveCardVersion.V1_5

    def test_config_notify_on_conversion(self):
        """Test notify_on string conversion."""
        from truthound.checkpoint.actions.teams_notify import TeamsConfig
        from truthound.checkpoint.actions.base import NotifyCondition

        config = TeamsConfig(
            webhook_url="https://test.webhook.office.com/test",
            notify_on="failure_or_error",
        )

        assert config.notify_on == NotifyCondition.FAILURE_OR_ERROR


# =============================================================================
# TeamsNotification Action Tests
# =============================================================================


class TestTeamsNotification:
    """Tests for TeamsNotification action class."""

    def test_default_config(self):
        """Test action has correct default config."""
        from truthound.checkpoint.actions.teams_notify import TeamsNotification

        action = TeamsNotification()
        assert action.action_type == "teams_notification"
        assert action.config.webhook_url == ""

    def test_config_via_kwargs(self):
        """Test config can be set via kwargs."""
        from truthound.checkpoint.actions.teams_notify import (
            TeamsNotification,
            MessageTheme,
        )

        action = TeamsNotification(
            webhook_url="https://test.webhook.office.com/test",
            theme=MessageTheme.MINIMAL,
            include_details=False,
        )

        assert action.config.webhook_url == "https://test.webhook.office.com/test"
        assert action.config.theme == MessageTheme.MINIMAL
        assert action.config.include_details is False

    def test_validate_config_missing_url(self):
        """Test validation fails without webhook URL."""
        from truthound.checkpoint.actions.teams_notify import TeamsNotification

        action = TeamsNotification()
        errors = action.validate_config()

        assert len(errors) == 1
        assert "webhook_url is required" in errors[0]

    def test_validate_config_invalid_url(self):
        """Test validation fails with invalid URL."""
        from truthound.checkpoint.actions.teams_notify import TeamsNotification

        action = TeamsNotification(webhook_url="https://invalid.example.com/webhook")
        errors = action.validate_config()

        assert len(errors) == 1
        assert "does not appear to be a valid Teams webhook URL" in errors[0]

    def test_validate_config_valid_webhook_url(self):
        """Test validation passes with valid webhook URL."""
        from truthound.checkpoint.actions.teams_notify import TeamsNotification

        # Test various valid URL patterns
        valid_urls = [
            "https://example.webhook.office.com/webhookb2/abc",
            "https://outlook.office.com/webhook/abc123",
            "https://prod-01.westus.logic.azure.com/workflows/abc",
        ]

        for url in valid_urls:
            action = TeamsNotification(webhook_url=url)
            errors = action.validate_config()
            assert errors == [], f"URL {url} should be valid"

    def test_execute_without_url(self, mock_checkpoint_result):
        """Test execute returns error without URL."""
        from truthound.checkpoint.actions.teams_notify import TeamsNotification
        from truthound.checkpoint.actions.base import ActionStatus

        action = TeamsNotification()
        result = action.execute(mock_checkpoint_result)

        assert result.status == ActionStatus.ERROR
        assert "No webhook URL configured" in result.message

    @patch("truthound.checkpoint.actions.teams_notify.TeamsHTTPClient")
    def test_execute_success(self, mock_client_class, mock_checkpoint_result):
        """Test successful execution."""
        from truthound.checkpoint.actions.teams_notify import TeamsNotification
        from truthound.checkpoint.actions.base import ActionStatus

        # Setup mock
        mock_client = MagicMock()
        mock_client.post.return_value = (200, "1")
        mock_client_class.return_value = mock_client

        action = TeamsNotification(
            webhook_url="https://test.webhook.office.com/test",
        )
        result = action.execute(mock_checkpoint_result)

        assert result.status == ActionStatus.SUCCESS
        assert "Teams notification sent" in result.message
        mock_client.post.assert_called_once()

    @patch("truthound.checkpoint.actions.teams_notify.TeamsHTTPClient")
    def test_execute_failure(self, mock_client_class, mock_checkpoint_result):
        """Test execution failure."""
        from truthound.checkpoint.actions.teams_notify import TeamsNotification
        from truthound.checkpoint.actions.base import ActionStatus

        mock_client = MagicMock()
        mock_client.post.return_value = (400, "Bad Request")
        mock_client_class.return_value = mock_client

        action = TeamsNotification(
            webhook_url="https://test.webhook.office.com/test",
        )
        result = action.execute(mock_checkpoint_result)

        assert result.status == ActionStatus.ERROR
        assert "400" in result.message

    @patch("truthound.checkpoint.actions.teams_notify.TeamsHTTPClient")
    def test_execute_exception(self, mock_client_class, mock_checkpoint_result):
        """Test execution with exception."""
        from truthound.checkpoint.actions.teams_notify import TeamsNotification
        from truthound.checkpoint.actions.base import ActionStatus

        mock_client = MagicMock()
        mock_client.post.side_effect = Exception("Network error")
        mock_client_class.return_value = mock_client

        action = TeamsNotification(
            webhook_url="https://test.webhook.office.com/test",
        )
        result = action.execute(mock_checkpoint_result)

        assert result.status == ActionStatus.ERROR
        assert "Network error" in result.error

    def test_execute_skipped_on_success(self, mock_success_result):
        """Test execution is skipped on success when notify_on=failure."""
        from truthound.checkpoint.actions.teams_notify import TeamsNotification
        from truthound.checkpoint.actions.base import ActionStatus

        action = TeamsNotification(
            webhook_url="https://test.webhook.office.com/test",
            notify_on="failure",
        )
        result = action.execute(mock_success_result)

        assert result.status == ActionStatus.SKIPPED

    @patch("truthound.checkpoint.actions.teams_notify.TeamsHTTPClient")
    def test_execute_with_custom_payload(self, mock_client_class, mock_checkpoint_result):
        """Test execution with custom payload."""
        from truthound.checkpoint.actions.teams_notify import TeamsNotification
        from truthound.checkpoint.actions.base import ActionStatus

        mock_client = MagicMock()
        mock_client.post.return_value = (200, "1")
        mock_client_class.return_value = mock_client

        custom_payload = {
            "type": "message",
            "text": "Checkpoint ${checkpoint} status: ${status}",
        }

        action = TeamsNotification(
            webhook_url="https://test.webhook.office.com/test",
            notify_on="always",
            custom_payload=custom_payload,
        )
        result = action.execute(mock_checkpoint_result)

        assert result.status == ActionStatus.SUCCESS

        # Verify placeholder substitution
        call_args = mock_client.post.call_args
        payload = call_args[0][1]
        assert "test_checkpoint" in payload["text"]
        assert "failure" in payload["text"]

    @patch("truthound.checkpoint.actions.teams_notify.TeamsHTTPClient")
    def test_execute_with_custom_template(self, mock_client_class, mock_checkpoint_result):
        """Test execution with custom template."""
        from truthound.checkpoint.actions.teams_notify import (
            TeamsNotification,
            MessageTemplate,
        )
        from truthound.checkpoint.actions.base import ActionStatus

        mock_client = MagicMock()
        mock_client.post.return_value = (200, "1")
        mock_client_class.return_value = mock_client

        class CustomTemplate(MessageTemplate):
            def render(self, checkpoint_result, config):
                return {
                    "type": "message",
                    "text": f"Custom: {checkpoint_result.checkpoint_name}",
                }

        action = TeamsNotification(
            webhook_url="https://test.webhook.office.com/test",
            notify_on="always",
            custom_template=CustomTemplate(),
        )
        result = action.execute(mock_checkpoint_result)

        assert result.status == ActionStatus.SUCCESS

        call_args = mock_client.post.call_args
        payload = call_args[0][1]
        assert payload["text"] == "Custom: test_checkpoint"


# =============================================================================
# HTTP Client Tests
# =============================================================================


class TestTeamsHTTPClient:
    """Tests for TeamsHTTPClient class."""

    @patch("urllib.request.urlopen")
    @patch("urllib.request.build_opener")
    def test_post_success(self, mock_build_opener, mock_urlopen):
        """Test successful POST request."""
        from truthound.checkpoint.actions.teams_notify import TeamsHTTPClient

        # Setup mock response
        mock_response = MagicMock()
        mock_response.getcode.return_value = 200
        mock_response.read.return_value = b"1"
        mock_response.__enter__ = MagicMock(return_value=mock_response)
        mock_response.__exit__ = MagicMock(return_value=False)

        mock_opener = MagicMock()
        mock_opener.open.return_value = mock_response
        mock_build_opener.return_value = mock_opener

        client = TeamsHTTPClient(timeout=30)
        status, body = client.post(
            "https://test.webhook.office.com/test",
            {"type": "message", "text": "Test"},
        )

        assert status == 200
        assert body == "1"

    @patch("urllib.request.urlopen")
    @patch("urllib.request.build_opener")
    def test_post_with_proxy(self, mock_build_opener, mock_urlopen):
        """Test POST request with proxy."""
        from truthound.checkpoint.actions.teams_notify import TeamsHTTPClient

        mock_response = MagicMock()
        mock_response.getcode.return_value = 200
        mock_response.read.return_value = b"1"
        mock_response.__enter__ = MagicMock(return_value=mock_response)
        mock_response.__exit__ = MagicMock(return_value=False)

        mock_opener = MagicMock()
        mock_opener.open.return_value = mock_response
        mock_build_opener.return_value = mock_opener

        client = TeamsHTTPClient(
            timeout=30,
            proxy="http://proxy.example.com:8080",
        )
        status, body = client.post(
            "https://test.webhook.office.com/test",
            {"type": "message"},
        )

        assert status == 200
        # Verify ProxyHandler was used
        mock_build_opener.assert_called()


# =============================================================================
# Convenience Function Tests
# =============================================================================


class TestConvenienceFunctions:
    """Tests for convenience functions."""

    def test_create_teams_notification(self):
        """Test create_teams_notification function."""
        from truthound.checkpoint.actions.teams_notify import (
            create_teams_notification,
            MessageTheme,
        )
        from truthound.checkpoint.actions.base import NotifyCondition

        action = create_teams_notification(
            "https://test.webhook.office.com/test",
            notify_on=NotifyCondition.FAILURE_OR_ERROR,
            theme=MessageTheme.DETAILED,
        )

        assert action.config.webhook_url == "https://test.webhook.office.com/test"
        assert action.config.notify_on == NotifyCondition.FAILURE_OR_ERROR
        assert action.config.theme == MessageTheme.DETAILED

    def test_create_failure_alert(self):
        """Test create_failure_alert function."""
        from truthound.checkpoint.actions.teams_notify import create_failure_alert
        from truthound.checkpoint.actions.base import NotifyCondition

        action = create_failure_alert(
            "https://test.webhook.office.com/test",
            mention_users=["user@example.com"],
            dashboard_url="https://dashboard.example.com",
        )

        assert action.config.notify_on == NotifyCondition.FAILURE_OR_ERROR
        assert action.config.mention_on_failure == ["user@example.com"]
        assert action.config.dashboard_url == "https://dashboard.example.com"
        assert action.config.include_actions is True

    def test_create_summary_notification(self):
        """Test create_summary_notification function."""
        from truthound.checkpoint.actions.teams_notify import (
            create_summary_notification,
            MessageTheme,
        )
        from truthound.checkpoint.actions.base import NotifyCondition

        action = create_summary_notification(
            "https://test.webhook.office.com/test",
            notify_on=NotifyCondition.ALWAYS,
        )

        assert action.config.theme == MessageTheme.COMPACT
        assert action.config.include_details is False
        assert action.config.include_actions is False


# =============================================================================
# Integration Tests
# =============================================================================


class TestTeamsNotificationIntegration:
    """Integration tests for Teams notification flow."""

    @patch("truthound.checkpoint.actions.teams_notify.TeamsHTTPClient")
    def test_full_notification_flow(self, mock_client_class, mock_checkpoint_result):
        """Test complete notification flow with all options."""
        from truthound.checkpoint.actions.teams_notify import (
            TeamsNotification,
            MessageTheme,
        )
        from truthound.checkpoint.actions.base import ActionStatus

        mock_client = MagicMock()
        mock_client.post.return_value = (200, "1")
        mock_client_class.return_value = mock_client

        action = TeamsNotification(
            webhook_url="https://test.webhook.office.com/test",
            notify_on="failure_or_error",
            theme=MessageTheme.DETAILED,
            full_width=True,
            include_details=True,
            include_actions=True,
            mention_on_failure=["user@example.com"],
            dashboard_url="https://dashboard.example.com/runs/{run_id}",
            details_url="https://details.example.com/{checkpoint}",
        )

        result = action.execute(mock_checkpoint_result)

        assert result.status == ActionStatus.SUCCESS

        # Verify payload structure
        call_args = mock_client.post.call_args
        payload = call_args[0][1]

        assert payload["type"] == "message"
        assert len(payload["attachments"]) == 1

        content = payload["attachments"][0]["content"]
        assert content["type"] == "AdaptiveCard"
        assert "msteams" in content
        assert content["msteams"].get("width") == "Full"

    @patch("truthound.checkpoint.actions.teams_notify.TeamsHTTPClient")
    def test_success_notification_skipped(self, mock_client_class, mock_success_result):
        """Test that success results are skipped with failure-only config."""
        from truthound.checkpoint.actions.teams_notify import TeamsNotification
        from truthound.checkpoint.actions.base import ActionStatus

        action = TeamsNotification(
            webhook_url="https://test.webhook.office.com/test",
            notify_on="failure",
        )

        result = action.execute(mock_success_result)

        assert result.status == ActionStatus.SKIPPED
        mock_client_class.assert_not_called()

    @patch("truthound.checkpoint.actions.teams_notify.TeamsHTTPClient")
    def test_success_notification_sent(self, mock_client_class, mock_success_result):
        """Test that success results are sent with always config."""
        from truthound.checkpoint.actions.teams_notify import TeamsNotification
        from truthound.checkpoint.actions.base import ActionStatus

        mock_client = MagicMock()
        mock_client.post.return_value = (200, "1")
        mock_client_class.return_value = mock_client

        action = TeamsNotification(
            webhook_url="https://test.webhook.office.com/test",
            notify_on="always",
        )

        result = action.execute(mock_success_result)

        assert result.status == ActionStatus.SUCCESS
        mock_client.post.assert_called_once()


# =============================================================================
# Module Export Tests
# =============================================================================


class TestModuleExports:
    """Tests for module exports."""

    def test_all_exports_accessible(self):
        """Test all expected exports are accessible from __init__."""
        from truthound.checkpoint.actions import (
            TeamsNotification,
            TeamsConfig,
            AdaptiveCardBuilder,
            MessageTemplate,
            MessageTheme,
            TeamsConnectorType,
            create_teams_notification,
            create_failure_alert,
            create_summary_notification,
            register_template,
        )

        # Just verify imports work
        assert TeamsNotification is not None
        assert TeamsConfig is not None
        assert AdaptiveCardBuilder is not None
        assert MessageTemplate is not None
        assert MessageTheme is not None
        assert TeamsConnectorType is not None
        assert create_teams_notification is not None
        assert create_failure_alert is not None
        assert create_summary_notification is not None
        assert register_template is not None

    def test_teams_notification_in_all(self):
        """Test TeamsNotification is in __all__."""
        from truthound.checkpoint import actions

        assert "TeamsNotification" in actions.__all__
        assert "TeamsConfig" in actions.__all__
        assert "AdaptiveCardBuilder" in actions.__all__



"""Execution engine implementations for Truthound.

This package provides execution engines that run validation operations
on different data backends.

Execution engines are responsible for:
- Running validation operations (count, aggregate, filter, etc.)
- Optimizing operations for the specific backend
- Converting to Polars for validators that require it
- Managing caching and performance

Available engines:
- PolarsExecutionEngine: Primary engine using Polars (default)
- PandasExecutionEngine: Native Pandas operations
- SQLExecutionEngine: SQL pushdown for database sources

Example:
    >>> from truthound.datasources import get_datasource
    >>> from truthound.execution import PolarsExecutionEngine
    >>>
    >>> source = get_datasource("data.csv")
    >>> engine = source.get_execution_engine()
    >>>
    >>> # Use engine for validation operations
    >>> null_count = engine.count_nulls("column_name")
    >>> stats = engine.get_stats("numeric_column")
"""

from truthound.execution._protocols import (
    AggregationType,
    ValidatorCapability,
    ExecutionEngineProtocol,
    SQLExecutionEngineProtocol,
)

from truthound.execution.base import (
    BaseExecutionEngine,
    ExecutionConfig,
    ExecutionError,
    UnsupportedOperationError,
    ExecutionSizeError,
)

from truthound.execution.polars_engine import PolarsExecutionEngine
from truthound.execution.pandas_engine import PandasExecutionEngine
from truthound.execution.sql_engine import SQLExecutionEngine

__all__ = [
    # Protocols
    "AggregationType",
    "ValidatorCapability",
    "ExecutionEngineProtocol",
    "SQLExecutionEngineProtocol",
    # Base classes
    "BaseExecutionEngine",
    "ExecutionConfig",
    # Exceptions
    "ExecutionError",
    "UnsupportedOperationError",
    "ExecutionSizeError",
    # Engines
    "PolarsExecutionEngine",
    "PandasExecutionEngine",
    "SQLExecutionEngine",
]



"""Validator-specific internationalized messages.

This module provides a complete i18n system for validator error messages,
supporting multiple languages and consistent formatting.
"""

from __future__ import annotations

import locale
import threading
from dataclasses import dataclass, field
from enum import Enum
from typing import Any


class ValidatorMessageCode(str, Enum):
    """Message codes for validator error messages.

    Codes are organized by validator category:
    - NULL: Null/missing value checks
    - UNIQUE: Uniqueness checks
    - TYPE: Data type checks
    - FORMAT: Format/pattern checks
    - RANGE: Value range checks
    - REFERENTIAL: Referential integrity checks
    - STATISTICAL: Statistical checks
    - SCHEMA: Schema validation
    - CUSTOM: Custom validators
    """

    # Null/Completeness checks
    NULL_VALUES_FOUND = "null.values_found"
    NULL_COLUMN_EMPTY = "null.column_empty"
    NULL_ABOVE_THRESHOLD = "null.above_threshold"

    # Uniqueness checks
    UNIQUE_DUPLICATES_FOUND = "unique.duplicates_found"
    UNIQUE_COMPOSITE_DUPLICATES = "unique.composite_duplicates"
    UNIQUE_KEY_VIOLATION = "unique.key_violation"

    # Type checks
    TYPE_MISMATCH = "type.mismatch"
    TYPE_COERCION_FAILED = "type.coercion_failed"
    TYPE_INFERENCE_FAILED = "type.inference_failed"

    # Format/Pattern checks
    FORMAT_INVALID_EMAIL = "format.invalid_email"
    FORMAT_INVALID_PHONE = "format.invalid_phone"
    FORMAT_INVALID_DATE = "format.invalid_date"
    FORMAT_INVALID_URL = "format.invalid_url"
    FORMAT_PATTERN_MISMATCH = "format.pattern_mismatch"
    FORMAT_REGEX_FAILED = "format.regex_failed"

    # Range checks
    RANGE_OUT_OF_BOUNDS = "range.out_of_bounds"
    RANGE_BELOW_MINIMUM = "range.below_minimum"
    RANGE_ABOVE_MAXIMUM = "range.above_maximum"
    RANGE_OUTLIER_DETECTED = "range.outlier_detected"

    # Referential integrity
    REF_FOREIGN_KEY_VIOLATION = "ref.foreign_key_violation"
    REF_MISSING_REFERENCE = "ref.missing_reference"
    REF_ORPHAN_RECORDS = "ref.orphan_records"

    # Statistical checks
    STAT_DISTRIBUTION_ANOMALY = "stat.distribution_anomaly"
    STAT_MEAN_OUT_OF_RANGE = "stat.mean_out_of_range"
    STAT_VARIANCE_ANOMALY = "stat.variance_anomaly"
    STAT_SKEWNESS_ANOMALY = "stat.skewness_anomaly"

    # Schema validation
    SCHEMA_COLUMN_MISSING = "schema.column_missing"
    SCHEMA_COLUMN_EXTRA = "schema.column_extra"
    SCHEMA_TYPE_MISMATCH = "schema.type_mismatch"
    SCHEMA_CONSTRAINT_VIOLATED = "schema.constraint_violated"

    # Cross-table checks
    CROSS_COLUMN_MISMATCH = "cross.column_mismatch"
    CROSS_CONSISTENCY_FAILED = "cross.consistency_failed"

    # Timeout/Performance
    TIMEOUT_EXCEEDED = "timeout.exceeded"
    TIMEOUT_PARTIAL_RESULT = "timeout.partial_result"

    # General
    VALIDATION_FAILED = "validation.failed"
    VALIDATION_SKIPPED = "validation.skipped"
    VALIDATION_ERROR = "validation.error"


# Default English messages
_DEFAULT_MESSAGES: dict[str, str] = {
    # Null/Completeness
    "null.values_found": "Found {count} null values in column '{column}'",
    "null.column_empty": "Column '{column}' is completely empty",
    "null.above_threshold": "Null ratio ({ratio:.1%}) exceeds threshold ({threshold:.1%}) in column '{column}'",

    # Uniqueness
    "unique.duplicates_found": "Found {count} duplicate values in column '{column}'",
    "unique.composite_duplicates": "Found {count} duplicate combinations for columns {columns}",
    "unique.key_violation": "Primary key violation: {count} duplicate keys in '{column}'",

    # Type
    "type.mismatch": "Type mismatch in column '{column}': expected {expected}, found {actual}",
    "type.coercion_failed": "Could not convert {count} values in column '{column}' to {target_type}",
    "type.inference_failed": "Could not infer type for column '{column}'",

    # Format
    "format.invalid_email": "Found {count} invalid email addresses in column '{column}'",
    "format.invalid_phone": "Found {count} invalid phone numbers in column '{column}'",
    "format.invalid_date": "Found {count} invalid dates in column '{column}'",
    "format.invalid_url": "Found {count} invalid URLs in column '{column}'",
    "format.pattern_mismatch": "Found {count} values not matching pattern '{pattern}' in column '{column}'",
    "format.regex_failed": "Regex validation failed for column '{column}': {count} non-matching values",

    # Range
    "range.out_of_bounds": "Found {count} values outside range [{min}, {max}] in column '{column}'",
    "range.below_minimum": "Found {count} values below minimum ({min}) in column '{column}'",
    "range.above_maximum": "Found {count} values above maximum ({max}) in column '{column}'",
    "range.outlier_detected": "Detected {count} statistical outliers in column '{column}'",

    # Referential
    "ref.foreign_key_violation": "Found {count} foreign key violations in column '{column}'",
    "ref.missing_reference": "Found {count} values in '{column}' not present in reference column '{ref_column}'",
    "ref.orphan_records": "Found {count} orphan records without matching parent in '{column}'",

    # Statistical
    "stat.distribution_anomaly": "Distribution anomaly detected in column '{column}': {details}",
    "stat.mean_out_of_range": "Mean value ({mean:.2f}) is outside expected range [{min}, {max}] for column '{column}'",
    "stat.variance_anomaly": "Unusual variance ({variance:.2f}) detected in column '{column}'",
    "stat.skewness_anomaly": "Skewness ({skewness:.2f}) indicates non-normal distribution in column '{column}'",

    # Schema
    "schema.column_missing": "Expected column '{column}' is missing from the dataset",
    "schema.column_extra": "Unexpected column '{column}' found in dataset",
    "schema.type_mismatch": "Column '{column}' has type {actual}, expected {expected}",
    "schema.constraint_violated": "Constraint '{constraint}' violated for column '{column}'",

    # Cross-table
    "cross.column_mismatch": "Values in column '{column1}' do not match column '{column2}'",
    "cross.consistency_failed": "Cross-column consistency check failed: {details}",

    # Timeout
    "timeout.exceeded": "Validation timed out after {seconds}s for '{operation}'",
    "timeout.partial_result": "Partial results returned due to timeout: validated {validated}% of data",

    # General
    "validation.failed": "Validation failed for column '{column}': {reason}",
    "validation.skipped": "Validation skipped for column '{column}': {reason}",
    "validation.error": "Error during validation: {error}",
}

# Korean messages
_KOREAN_MESSAGES: dict[str, str] = {
    # Null/완전성
    "null.values_found": "'{column}' 컬럼에서 {count}개의 null 값이 발견되었습니다",
    "null.column_empty": "'{column}' 컬럼이 완전히 비어있습니다",
    "null.above_threshold": "'{column}' 컬럼의 null 비율({ratio:.1%})이 임계값({threshold:.1%})을 초과했습니다",

    # 고유성
    "unique.duplicates_found": "'{column}' 컬럼에서 {count}개의 중복 값이 발견되었습니다",
    "unique.composite_duplicates": "{columns} 컬럼에서 {count}개의 중복 조합이 발견되었습니다",
    "unique.key_violation": "기본키 위반: '{column}'에서 {count}개의 중복 키가 발견되었습니다",

    # 타입
    "type.mismatch": "'{column}' 컬럼의 타입 불일치: {expected} 예상, {actual} 발견",
    "type.coercion_failed": "'{column}' 컬럼에서 {count}개의 값을 {target_type}으로 변환할 수 없습니다",
    "type.inference_failed": "'{column}' 컬럼의 타입을 추론할 수 없습니다",

    # 형식
    "format.invalid_email": "'{column}' 컬럼에서 {count}개의 잘못된 이메일 주소가 발견되었습니다",
    "format.invalid_phone": "'{column}' 컬럼에서 {count}개의 잘못된 전화번호가 발견되었습니다",
    "format.invalid_date": "'{column}' 컬럼에서 {count}개의 잘못된 날짜가 발견되었습니다",
    "format.invalid_url": "'{column}' 컬럼에서 {count}개의 잘못된 URL이 발견되었습니다",
    "format.pattern_mismatch": "'{column}' 컬럼에서 패턴 '{pattern}'과 일치하지 않는 {count}개의 값이 발견되었습니다",
    "format.regex_failed": "'{column}' 컬럼의 정규식 검증 실패: {count}개의 불일치 값",

    # 범위
    "range.out_of_bounds": "'{column}' 컬럼에서 범위 [{min}, {max}]를 벗어난 {count}개의 값이 발견되었습니다",
    "range.below_minimum": "'{column}' 컬럼에서 최소값({min}) 미만인 {count}개의 값이 발견되었습니다",
    "range.above_maximum": "'{column}' 컬럼에서 최대값({max}) 초과인 {count}개의 값이 발견되었습니다",
    "range.outlier_detected": "'{column}' 컬럼에서 {count}개의 통계적 이상치가 감지되었습니다",

    # 참조 무결성
    "ref.foreign_key_violation": "'{column}' 컬럼에서 {count}개의 외래키 위반이 발견되었습니다",
    "ref.missing_reference": "'{column}'의 {count}개 값이 참조 컬럼 '{ref_column}'에 존재하지 않습니다",
    "ref.orphan_records": "'{column}'에서 상위 레코드와 일치하지 않는 {count}개의 고아 레코드가 발견되었습니다",

    # 통계
    "stat.distribution_anomaly": "'{column}' 컬럼에서 분포 이상이 감지되었습니다: {details}",
    "stat.mean_out_of_range": "'{column}' 컬럼의 평균값({mean:.2f})이 예상 범위 [{min}, {max}]를 벗어났습니다",
    "stat.variance_anomaly": "'{column}' 컬럼에서 비정상적인 분산({variance:.2f})이 감지되었습니다",
    "stat.skewness_anomaly": "'{column}' 컬럼의 왜도({skewness:.2f})가 비정규 분포를 나타냅니다",

    # 스키마
    "schema.column_missing": "예상 컬럼 '{column}'이(가) 데이터셋에서 누락되었습니다",
    "schema.column_extra": "예상치 못한 컬럼 '{column}'이(가) 데이터셋에서 발견되었습니다",
    "schema.type_mismatch": "'{column}' 컬럼의 타입이 {actual}이지만, {expected}이어야 합니다",
    "schema.constraint_violated": "'{column}' 컬럼에서 제약조건 '{constraint}'이(가) 위반되었습니다",

    # 교차 테이블
    "cross.column_mismatch": "'{column1}' 컬럼의 값이 '{column2}' 컬럼과 일치하지 않습니다",
    "cross.consistency_failed": "교차 컬럼 일관성 검사 실패: {details}",

    # 타임아웃
    "timeout.exceeded": "'{operation}'에 대한 검증이 {seconds}초 후 타임아웃되었습니다",
    "timeout.partial_result": "타임아웃으로 인해 부분 결과 반환: 데이터의 {validated}% 검증됨",

    # 일반
    "validation.failed": "'{column}' 컬럼 검증 실패: {reason}",
    "validation.skipped": "'{column}' 컬럼 검증 건너뜀: {reason}",
    "validation.error": "검증 중 오류 발생: {error}",
}

# Japanese messages
_JAPANESE_MESSAGES: dict[str, str] = {
    "null.values_found": "列'{column}'で{count}個のnull値が見つかりました",
    "null.column_empty": "列'{column}'は完全に空です",
    "null.above_threshold": "列'{column}'のnull比率({ratio:.1%})が閾値({threshold:.1%})を超えています",

    "unique.duplicates_found": "列'{column}'で{count}個の重複値が見つかりました",
    "unique.composite_duplicates": "列{columns}で{count}個の重複組み合わせが見つかりました",
    "unique.key_violation": "主キー違反: '{column}'で{count}個の重複キーが見つかりました",

    "type.mismatch": "列'{column}'の型不一致: {expected}を期待、{actual}が見つかりました",
    "type.coercion_failed": "列'{column}'の{count}個の値を{target_type}に変換できません",
    "type.inference_failed": "列'{column}'の型を推論できません",

    "format.invalid_email": "列'{column}'で{count}個の無効なメールアドレスが見つかりました",
    "format.invalid_phone": "列'{column}'で{count}個の無効な電話番号が見つかりました",
    "format.invalid_date": "列'{column}'で{count}個の無効な日付が見つかりました",
    "format.invalid_url": "列'{column}'で{count}個の無効なURLが見つかりました",
    "format.pattern_mismatch": "列'{column}'でパターン'{pattern}'に一致しない{count}個の値が見つかりました",
    "format.regex_failed": "列'{column}'の正規表現検証失敗: {count}個の不一致値",

    "range.out_of_bounds": "列'{column}'で範囲[{min}, {max}]外の{count}個の値が見つかりました",
    "range.below_minimum": "列'{column}'で最小値({min})未満の{count}個の値が見つかりました",
    "range.above_maximum": "列'{column}'で最大値({max})超過の{count}個の値が見つかりました",
    "range.outlier_detected": "列'{column}'で{count}個の統計的外れ値が検出されました",

    "ref.foreign_key_violation": "列'{column}'で{count}個の外部キー違反が見つかりました",
    "ref.missing_reference": "'{column}'の{count}個の値が参照列'{ref_column}'に存在しません",
    "ref.orphan_records": "'{column}'で親レコードと一致しない{count}個の孤立レコードが見つかりました",

    "stat.distribution_anomaly": "列'{column}'で分布異常が検出されました: {details}",
    "stat.mean_out_of_range": "列'{column}'の平均値({mean:.2f})が予想範囲[{min}, {max}]外です",
    "stat.variance_anomaly": "列'{column}'で異常な分散({variance:.2f})が検出されました",
    "stat.skewness_anomaly": "列'{column}'の歪度({skewness:.2f})は非正規分布を示しています",

    "schema.column_missing": "期待される列'{column}'がデータセットにありません",
    "schema.column_extra": "予期しない列'{column}'がデータセットに見つかりました",
    "schema.type_mismatch": "列'{column}'の型は{actual}ですが、{expected}であるべきです",
    "schema.constraint_violated": "列'{column}'で制約'{constraint}'が違反されました",

    "cross.column_mismatch": "列'{column1}'の値が列'{column2}'と一致しません",
    "cross.consistency_failed": "クロス列一貫性チェック失敗: {details}",

    "timeout.exceeded": "'{operation}'の検証が{seconds}秒後にタイムアウトしました",
    "timeout.partial_result": "タイムアウトにより部分結果を返却: データの{validated}%を検証",

    "validation.failed": "列'{column}'の検証失敗: {reason}",
    "validation.skipped": "列'{column}'の検証スキップ: {reason}",
    "validation.error": "検証中にエラーが発生しました: {error}",
}

# Chinese (Simplified) messages
_CHINESE_MESSAGES: dict[str, str] = {
    "null.values_found": "在列'{column}'中发现了{count}个空值",
    "null.column_empty": "列'{column}'完全为空",
    "null.above_threshold": "列'{column}'的空值比例({ratio:.1%})超过了阈值({threshold:.1%})",

    "unique.duplicates_found": "在列'{column}'中发现了{count}个重复值",
    "unique.composite_duplicates": "在列{columns}中发现了{count}个重复组合",
    "unique.key_violation": "主键违规: 在'{column}'中发现了{count}个重复键",

    "type.mismatch": "列'{column}'类型不匹配: 期望{expected}，发现{actual}",
    "type.coercion_failed": "无法将列'{column}'中的{count}个值转换为{target_type}",
    "type.inference_failed": "无法推断列'{column}'的类型",

    "format.invalid_email": "在列'{column}'中发现了{count}个无效的电子邮件地址",
    "format.invalid_phone": "在列'{column}'中发现了{count}个无效的电话号码",
    "format.invalid_date": "在列'{column}'中发现了{count}个无效的日期",
    "format.invalid_url": "在列'{column}'中发现了{count}个无效的URL",
    "format.pattern_mismatch": "在列'{column}'中发现了{count}个不匹配模式'{pattern}'的值",
    "format.regex_failed": "列'{column}'的正则表达式验证失败: {count}个不匹配的值",

    "range.out_of_bounds": "在列'{column}'中发现了{count}个超出范围[{min}, {max}]的值",
    "range.below_minimum": "在列'{column}'中发现了{count}个低于最小值({min})的值",
    "range.above_maximum": "在列'{column}'中发现了{count}个超过最大值({max})的值",
    "range.outlier_detected": "在列'{column}'中检测到{count}个统计异常值",

    "validation.failed": "列'{column}'验证失败: {reason}",
    "validation.skipped": "跳过列'{column}'的验证: {reason}",
    "validation.error": "验证过程中发生错误: {error}",
}

# German messages
_GERMAN_MESSAGES: dict[str, str] = {
    "null.values_found": "{count} Nullwerte in Spalte '{column}' gefunden",
    "null.column_empty": "Spalte '{column}' ist vollständig leer",
    "null.above_threshold": "Nullanteil ({ratio:.1%}) überschreitet Schwellenwert ({threshold:.1%}) in Spalte '{column}'",

    "unique.duplicates_found": "{count} Duplikate in Spalte '{column}' gefunden",
    "unique.key_violation": "Primärschlüsselverletzung: {count} doppelte Schlüssel in '{column}'",

    "type.mismatch": "Typabweichung in Spalte '{column}': erwartet {expected}, gefunden {actual}",
    "format.pattern_mismatch": "{count} Werte in Spalte '{column}' entsprechen nicht dem Muster '{pattern}'",

    "validation.failed": "Validierung für Spalte '{column}' fehlgeschlagen: {reason}",
    "validation.error": "Fehler während der Validierung: {error}",
}

# French messages
_FRENCH_MESSAGES: dict[str, str] = {
    "null.values_found": "{count} valeurs nulles trouvées dans la colonne '{column}'",
    "null.column_empty": "La colonne '{column}' est entièrement vide",

    "unique.duplicates_found": "{count} valeurs en double trouvées dans la colonne '{column}'",
    "type.mismatch": "Incompatibilité de type dans la colonne '{column}': attendu {expected}, trouvé {actual}",

    "validation.failed": "Échec de la validation pour la colonne '{column}': {reason}",
    "validation.error": "Erreur lors de la validation: {error}",
}

# Spanish messages
_SPANISH_MESSAGES: dict[str, str] = {
    "null.values_found": "Se encontraron {count} valores nulos en la columna '{column}'",
    "null.column_empty": "La columna '{column}' está completamente vacía",

    "unique.duplicates_found": "Se encontraron {count} valores duplicados en la columna '{column}'",
    "type.mismatch": "Discrepancia de tipo en la columna '{column}': esperado {expected}, encontrado {actual}",

    "validation.failed": "Falló la validación para la columna '{column}': {reason}",
    "validation.error": "Error durante la validación: {error}",
}


@dataclass
class ValidatorI18n:
    """Internationalization manager for validator messages.

    Thread-safe singleton that manages locale and message resolution.

    Example:
        i18n = ValidatorI18n.get_instance()
        i18n.set_locale("ko")
        msg = i18n.get_message(ValidatorMessageCode.NULL_VALUES_FOUND, column="email", count=5)
    """

    _locale: str = "en"
    _fallback_locale: str = "en"
    _catalogs: dict[str, dict[str, str]] = field(default_factory=dict)
    _lock: threading.Lock = field(default_factory=threading.Lock)

    _instance: "ValidatorI18n | None" = None

    @classmethod
    def get_instance(cls) -> "ValidatorI18n":
        """Get singleton instance."""
        if cls._instance is None:
            cls._instance = cls()
            cls._instance._initialize_catalogs()
        return cls._instance

    def _initialize_catalogs(self) -> None:
        """Initialize message catalogs."""
        self._catalogs = {
            "en": _DEFAULT_MESSAGES,
            "ko": _KOREAN_MESSAGES,
            "ja": _JAPANESE_MESSAGES,
            "zh": _CHINESE_MESSAGES,
            "de": _GERMAN_MESSAGES,
            "fr": _FRENCH_MESSAGES,
            "es": _SPANISH_MESSAGES,
        }

    def set_locale(self, locale_code: str) -> None:
        """Set the current locale.

        Args:
            locale_code: Locale code (e.g., "en", "ko", "ja")
        """
        with self._lock:
            # Extract language code (e.g., "ko_KR" -> "ko")
            lang = locale_code.split("_")[0].lower()
            if lang in self._catalogs:
                self._locale = lang
            else:
                self._locale = self._fallback_locale

    def get_locale(self) -> str:
        """Get the current locale.

        Returns:
            Current locale code
        """
        return self._locale

    def get_message(
        self,
        code: ValidatorMessageCode | str,
        **kwargs: Any,
    ) -> str:
        """Get a localized message.

        Args:
            code: Message code
            **kwargs: Format arguments

        Returns:
            Formatted localized message
        """
        key = code.value if isinstance(code, ValidatorMessageCode) else code

        # Try current locale
        catalog = self._catalogs.get(self._locale, {})
        template = catalog.get(key)

        # Fallback to English
        if template is None:
            catalog = self._catalogs.get(self._fallback_locale, {})
            template = catalog.get(key, f"[{key}]")

        # Format message
        try:
            return template.format(**kwargs)
        except KeyError as e:
            # Missing placeholder
            return f"{template} (missing: {e})"
        except Exception:
            return template

    def add_catalog(self, locale_code: str, messages: dict[str, str]) -> None:
        """Add or update a message catalog.

        Args:
            locale_code: Locale code
            messages: Dictionary of message key -> template
        """
        with self._lock:
            lang = locale_code.split("_")[0].lower()
            if lang in self._catalogs:
                self._catalogs[lang].update(messages)
            else:
                self._catalogs[lang] = messages


# Module-level convenience functions

def get_validator_message(
    code: ValidatorMessageCode | str,
    **kwargs: Any,
) -> str:
    """Get a localized validator message.

    Args:
        code: Message code
        **kwargs: Format arguments

    Returns:
        Formatted message
    """
    return ValidatorI18n.get_instance().get_message(code, **kwargs)


def set_validator_locale(locale_code: str) -> None:
    """Set the validator locale.

    Args:
        locale_code: Locale code (e.g., "en", "ko", "ja")
    """
    ValidatorI18n.get_instance().set_locale(locale_code)


def get_validator_locale() -> str:
    """Get the current validator locale.

    Returns:
        Current locale code
    """
    return ValidatorI18n.get_instance().get_locale()


def format_issue_message(
    issue_type: str,
    column: str,
    count: int,
    **extra: Any,
) -> str:
    """Format a validation issue message.

    This function maps common issue types to their message codes
    and formats them with the provided arguments.

    Args:
        issue_type: Issue type (e.g., "null", "duplicate", "format")
        column: Column name
        count: Number of issues
        **extra: Additional format arguments

    Returns:
        Formatted message
    """
    # Map issue types to message codes
    type_to_code = {
        "null": ValidatorMessageCode.NULL_VALUES_FOUND,
        "null_value": ValidatorMessageCode.NULL_VALUES_FOUND,
        "duplicate": ValidatorMessageCode.UNIQUE_DUPLICATES_FOUND,
        "type_mismatch": ValidatorMessageCode.TYPE_MISMATCH,
        "invalid_email": ValidatorMessageCode.FORMAT_INVALID_EMAIL,
        "invalid_phone": ValidatorMessageCode.FORMAT_INVALID_PHONE,
        "invalid_date": ValidatorMessageCode.FORMAT_INVALID_DATE,
        "pattern_mismatch": ValidatorMessageCode.FORMAT_PATTERN_MISMATCH,
        "out_of_range": ValidatorMessageCode.RANGE_OUT_OF_BOUNDS,
        "outlier": ValidatorMessageCode.RANGE_OUTLIER_DETECTED,
        "foreign_key_violation": ValidatorMessageCode.REF_FOREIGN_KEY_VIOLATION,
        "schema_mismatch": ValidatorMessageCode.SCHEMA_TYPE_MISMATCH,
    }

    code = type_to_code.get(issue_type.lower(), ValidatorMessageCode.VALIDATION_FAILED)
    return get_validator_message(code, column=column, count=count, **extra)



"""Tests for RayExecutionEngine.

This module tests the Ray-native distributed execution engine.
"""

import pytest
from unittest.mock import MagicMock, patch, PropertyMock
from dataclasses import dataclass
from typing import Any


# =============================================================================
# Test Fixtures
# =============================================================================


@pytest.fixture
def mock_ray_available():
    """Mock ray availability check."""
    with patch(
        "truthound.execution.distributed.ray_engine._check_ray_available"
    ) as check_mock, patch(
        "truthound.execution.distributed.ray_engine._ensure_ray_initialized"
    ) as init_mock:
        yield check_mock, init_mock


@pytest.fixture
def mock_ray_dataset():
    """Create a mock Ray Dataset."""
    mock_ds = MagicMock()

    # Schema mock
    mock_schema = MagicMock()
    mock_schema.names = ["id", "name", "value", "category"]
    mock_ds.schema.return_value = mock_schema

    # Basic operations
    mock_ds.count.return_value = 1000
    mock_ds.num_blocks.return_value = 4

    return mock_ds


@pytest.fixture
def sample_data():
    """Create sample data for testing."""
    return [
        {"id": 1, "name": "Alice", "value": 100.0, "category": "A"},
        {"id": 2, "name": "Bob", "value": 200.0, "category": "B"},
        {"id": 3, "name": None, "value": 150.0, "category": "A"},
        {"id": 4, "name": "David", "value": None, "category": "B"},
        {"id": 5, "name": "Eve", "value": 300.0, "category": "A"},
        {"id": 6, "name": "Frank", "value": 250.0, "category": "B"},
        {"id": 7, "name": None, "value": 175.0, "category": "A"},
        {"id": 8, "name": "Heidi", "value": None, "category": "B"},
        {"id": 9, "name": "Ivan", "value": 225.0, "category": "A"},
        {"id": 10, "name": "Judy", "value": 275.0, "category": "B"},
    ]


# =============================================================================
# RayEngineConfig Tests
# =============================================================================


class TestRayEngineConfig:
    """Tests for RayEngineConfig."""

    def test_default_config(self):
        """Test default configuration values."""
        from truthound.execution.distributed.ray_engine import RayEngineConfig

        config = RayEngineConfig()

        assert config.ray_address is None
        assert config.num_cpus is None
        assert config.num_gpus is None
        assert config.batch_size == 4096
        assert config.prefetch_batches == 2
        assert config.concurrency is None
        assert config.use_actors is False
        assert config.actor_pool_size == 4
        assert config.target_max_block_size == 128 * 1024 * 1024

    def test_custom_config(self):
        """Test custom configuration."""
        from truthound.execution.distributed.ray_engine import RayEngineConfig

        config = RayEngineConfig(
            ray_address="ray://cluster:10001",
            num_cpus=16,
            num_gpus=2,
            batch_size=8192,
            use_actors=True,
            actor_pool_size=8,
        )

        assert config.ray_address == "ray://cluster:10001"
        assert config.num_cpus == 16
        assert config.num_gpus == 2
        assert config.batch_size == 8192
        assert config.use_actors is True
        assert config.actor_pool_size == 8


# =============================================================================
# RayExecutionEngine Core Tests
# =============================================================================


class TestRayExecutionEngineCore:
    """Core functionality tests for RayExecutionEngine."""

    def test_backend_type(self, mock_ray_available, mock_ray_dataset):
        """Test backend type property."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine
        from truthound.execution.distributed.protocols import ComputeBackend

        engine = RayExecutionEngine(mock_ray_dataset)

        assert engine.backend_type == ComputeBackend.RAY

    def test_engine_type(self, mock_ray_available, mock_ray_dataset):
        """Test engine type attribute."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine

        engine = RayExecutionEngine(mock_ray_dataset)

        assert engine.engine_type == "ray"

    def test_get_columns(self, mock_ray_available, mock_ray_dataset):
        """Test getting column names."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine

        engine = RayExecutionEngine(mock_ray_dataset)

        columns = engine.get_columns()

        assert columns == ["id", "name", "value", "category"]

    def test_get_partition_count(self, mock_ray_available, mock_ray_dataset):
        """Test getting partition (block) count."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine

        engine = RayExecutionEngine(mock_ray_dataset)

        assert engine._get_partition_count() == 4
        assert engine.num_partitions == 4

    def test_get_partition_info(self, mock_ray_available, mock_ray_dataset):
        """Test getting partition information."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine

        engine = RayExecutionEngine(mock_ray_dataset)

        infos = engine._get_partition_info()

        assert len(infos) == 4
        for i, info in enumerate(infos):
            assert info.partition_id == i
            assert info.total_partitions == 4
            assert info.columns == ("id", "name", "value", "category")


# =============================================================================
# RayExecutionEngine Factory Method Tests
# =============================================================================


class TestRayExecutionEngineFactoryMethods:
    """Tests for RayExecutionEngine factory methods."""

    def test_from_dataset(self, mock_ray_available, mock_ray_dataset):
        """Test creating engine from Dataset."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine

        engine = RayExecutionEngine.from_dataset(mock_ray_dataset)

        assert engine.dataset is mock_ray_dataset

    def test_from_dataset_with_config(self, mock_ray_available, mock_ray_dataset):
        """Test creating engine with custom config."""
        from truthound.execution.distributed.ray_engine import (
            RayExecutionEngine,
            RayEngineConfig,
        )

        config = RayEngineConfig(
            batch_size=8192,
            prefetch_batches=4,
        )

        engine = RayExecutionEngine.from_dataset(mock_ray_dataset, config=config)

        assert engine._config.batch_size == 8192
        assert engine._config.prefetch_batches == 4


# =============================================================================
# RayExecutionEngine Aggregation Tests
# =============================================================================


class TestRayExecutionEngineAggregations:
    """Tests for aggregation operations."""

    def test_count_rows(self, mock_ray_available, mock_ray_dataset):
        """Test counting rows."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine

        engine = RayExecutionEngine(mock_ray_dataset)
        count = engine.count_rows()

        assert count == 1000

    def test_count_rows_caching(self, mock_ray_available, mock_ray_dataset):
        """Test that count_rows is cached."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine

        engine = RayExecutionEngine(mock_ray_dataset)

        count1 = engine.count_rows()
        count2 = engine.count_rows()

        assert count1 == count2
        # count() should only be called once due to caching
        assert mock_ray_dataset.count.call_count == 1


# =============================================================================
# RayExecutionEngine Context Manager Tests
# =============================================================================


class TestRayExecutionEngineContextManager:
    """Tests for context manager functionality."""

    def test_context_manager_enter_exit(self, mock_ray_available, mock_ray_dataset):
        """Test context manager protocol."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine

        engine = RayExecutionEngine(mock_ray_dataset)

        with engine as e:
            assert e is engine


# =============================================================================
# Integration Tests (require ray)
# =============================================================================


@pytest.mark.skipif(
    not pytest.importorskip("ray", reason="ray not installed"),
    reason="ray not available",
)
class TestRayExecutionEngineIntegration:
    """Integration tests requiring actual ray."""

    @pytest.fixture(autouse=True)
    def setup_ray(self):
        """Initialize ray for tests."""
        import ray
        if not ray.is_initialized():
            ray.init(num_cpus=2, ignore_reinit_error=True)
        yield
        # Don't shutdown ray between tests

    def test_from_items_integration(self, sample_data):
        """Test creating engine from list of items."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine

        engine = RayExecutionEngine.from_items(sample_data)

        assert engine.count_rows() == 10
        assert set(engine.get_columns()) == {"id", "name", "value", "category"}

    def test_count_rows_integration(self, sample_data):
        """Test counting rows with actual data."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine

        engine = RayExecutionEngine.from_items(sample_data)

        assert engine.count_rows() == 10

    def test_sum_integration(self, sample_data):
        """Test sum aggregation."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine

        engine = RayExecutionEngine.from_items(sample_data)

        total = engine.dataset.sum("id")

        assert total == 55  # 1 + 2 + ... + 10

    def test_mean_integration(self, sample_data):
        """Test mean aggregation."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine

        engine = RayExecutionEngine.from_items(sample_data)

        mean = engine.dataset.mean("id")

        assert mean == 5.5

    def test_min_max_integration(self, sample_data):
        """Test min/max aggregation."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine

        engine = RayExecutionEngine.from_items(sample_data)

        min_val = engine.dataset.min("id")
        max_val = engine.dataset.max("id")

        assert min_val == 1
        assert max_val == 10

    def test_sample_integration(self, sample_data):
        """Test sampling."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine

        engine = RayExecutionEngine.from_items(sample_data)

        sampled = engine.sample(n=5, seed=42)

        assert sampled.count_rows() <= 5

    def test_repartition_integration(self, sample_data):
        """Test repartitioning."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine

        engine = RayExecutionEngine.from_items(sample_data)
        original_blocks = engine.num_partitions

        new_engine = engine.repartition(4)

        assert new_engine.num_partitions == 4
        assert new_engine.count_rows() == 10

    def test_take_integration(self, sample_data):
        """Test take operation."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine

        engine = RayExecutionEngine.from_items(sample_data)

        rows = engine.take(3)

        assert len(rows) == 3
        assert all(isinstance(r, dict) for r in rows)

    def test_to_pandas_integration(self, sample_data):
        """Test conversion to pandas."""
        pytest.importorskip("pandas")
        from truthound.execution.distributed.ray_engine import RayExecutionEngine
        import pandas as pd

        engine = RayExecutionEngine.from_items(sample_data)

        df = engine.to_pandas()

        assert isinstance(df, pd.DataFrame)
        assert len(df) == 10

    def test_to_arrow_integration(self, sample_data):
        """Test conversion to Arrow."""
        pytest.importorskip("pyarrow")
        from truthound.execution.distributed.ray_engine import RayExecutionEngine
        import pyarrow as pa

        engine = RayExecutionEngine.from_items(sample_data)

        table = engine.to_arrow()

        assert isinstance(table, pa.Table)
        assert table.num_rows == 10

    def test_to_polars_lazyframe_integration(self, sample_data):
        """Test conversion to Polars LazyFrame."""
        pytest.importorskip("polars")
        pytest.importorskip("pyarrow")

        from truthound.execution.distributed.ray_engine import RayExecutionEngine
        import polars as pl

        engine = RayExecutionEngine.from_items(sample_data)

        lf = engine.to_polars_lazyframe()

        assert isinstance(lf, pl.LazyFrame)
        df = lf.collect()
        assert len(df) == 10

    def test_materialize_integration(self, sample_data):
        """Test materializing the dataset."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine

        engine = RayExecutionEngine.from_items(sample_data)

        materialized = engine.materialize()

        assert materialized.count_rows() == 10

    def test_select_columns_integration(self, sample_data):
        """Test column selection."""
        from truthound.execution.distributed.ray_engine import RayExecutionEngine

        engine = RayExecutionEngine.from_items(sample_data)

        selected = engine.select_columns(["id", "name"])

        assert set(selected.get_columns()) == {"id", "name"}


# =============================================================================
# Registry Integration Tests
# =============================================================================


class TestRayEngineRegistry:
    """Tests for ray engine registration in the global registry."""

    def test_ray_engine_registered(self):
        """Test that ray engine is registered."""
        from truthound.execution.distributed.registry import list_distributed_engines

        engines = list_distributed_engines()

        assert "ray" in engines

    def test_auto_detect_ray_dataset(self, mock_ray_available, mock_ray_dataset):
        """Test auto-detection of Ray Dataset."""
        from truthound.execution.distributed.registry import (
            get_engine_registry,
        )
        from truthound.execution.distributed.protocols import ComputeBackend

        # Mock the module detection
        type(mock_ray_dataset).__module__ = PropertyMock(return_value="ray.data")
        type(mock_ray_dataset).__name__ = PropertyMock(return_value="Dataset")

        registry = get_engine_registry()
        backend = registry._detect_backend(mock_ray_dataset)

        assert backend == ComputeBackend.RAY



"""Integration tests for Kinesis adapter using LocalStack.

These tests require Docker to be running and testcontainers installed.
Run with: pytest tests/integration/realtime/test_kinesis_adapter.py -v
"""

from __future__ import annotations

import asyncio
import pytest
from datetime import datetime, timezone

pytest_plugins = ["pytest_asyncio"]

try:
    import testcontainers
    TESTCONTAINERS_AVAILABLE = True
except ImportError:
    TESTCONTAINERS_AVAILABLE = False

try:
    import aiobotocore
    AIOBOTOCORE_AVAILABLE = True
except ImportError:
    AIOBOTOCORE_AVAILABLE = False


@pytest.fixture(scope="module")
def event_loop():
    """Create event loop for async tests."""
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.mark.skipif(
    not TESTCONTAINERS_AVAILABLE,
    reason="testcontainers not installed"
)
@pytest.mark.skipif(
    not AIOBOTOCORE_AVAILABLE,
    reason="aiobotocore not installed"
)
class TestKinesisAdapterIntegration:
    """Integration tests for KinesisAdapter with LocalStack."""

    @pytest.fixture(scope="class")
    async def localstack_container(self):
        """Start LocalStack container for tests."""
        from truthound.realtime.testing.containers import LocalStackTestContainer

        async with LocalStackTestContainer() as localstack:
            # Create test stream
            await localstack.create_kinesis_stream("test-stream", shard_count=2)
            yield localstack

    @pytest.fixture
    def kinesis_config(self, localstack_container):
        """Get Kinesis adapter configuration."""
        from truthound.realtime.adapters.kinesis import KinesisAdapterConfig

        return KinesisAdapterConfig(
            stream_name="test-stream",
            region_name="us-east-1",
            endpoint_url=localstack_container.get_connection_url(),
            aws_access_key_id="test",
            aws_secret_access_key="test",
        )

    @pytest.mark.asyncio
    async def test_connect_and_disconnect(self, kinesis_config):
        """Test connecting and disconnecting from Kinesis."""
        from truthound.realtime.adapters.kinesis import KinesisAdapter

        adapter = KinesisAdapter(kinesis_config)

        # Connect
        await adapter.connect()
        assert adapter.is_connected

        # Disconnect
        await adapter.disconnect()
        assert not adapter.is_connected

    @pytest.mark.asyncio
    async def test_produce_and_consume(self, kinesis_config):
        """Test producing and consuming messages."""
        from truthound.realtime.adapters.kinesis import KinesisAdapter
        from truthound.realtime.protocols import StreamMessage

        adapter = KinesisAdapter(kinesis_config)

        async with adapter:
            # Produce messages
            for i in range(10):
                message = StreamMessage(
                    key=f"partition-key-{i % 3}",  # Distribute across shards
                    value={"id": i, "data": f"kinesis-test-{i}"},
                    partition=-1,
                    offset=-1,
                    timestamp=datetime.now(timezone.utc),
                    headers=(),
                    topic="test-stream",
                )
                await adapter.produce(message)

            await adapter.flush()

            # Wait for messages to be available
            await asyncio.sleep(2)

            # Consume messages
            consumed = []
            timeout = asyncio.get_event_loop().time() + 30  # 30 second timeout

            async for msg in adapter.consume():
                consumed.append(msg)

                if len(consumed) >= 10:
                    break
                if asyncio.get_event_loop().time() > timeout:
                    break

            assert len(consumed) >= 1  # At least some messages consumed

    @pytest.mark.asyncio
    async def test_batch_produce(self, kinesis_config):
        """Test batch message production."""
        from truthound.realtime.adapters.kinesis import KinesisAdapter
        from truthound.realtime.protocols import StreamMessage

        adapter = KinesisAdapter(kinesis_config)

        async with adapter:
            messages = [
                StreamMessage(
                    key=f"batch-pk-{i % 5}",
                    value={"batch_id": i, "data": f"batch-{i}"},
                    partition=-1,
                    offset=-1,
                    timestamp=datetime.now(timezone.utc),
                    headers=(),
                    topic="test-stream",
                )
                for i in range(50)
            ]

            await adapter.produce_batch(messages)
            await adapter.flush()

            # Verify metrics
            metrics = adapter.metrics
            assert metrics.messages_produced >= 50

    @pytest.mark.asyncio
    async def test_shard_info(self, kinesis_config):
        """Test getting shard information."""
        from truthound.realtime.adapters.kinesis import KinesisAdapter

        adapter = KinesisAdapter(kinesis_config)

        async with adapter:
            shards = await adapter.get_shard_info()
            assert len(shards) == 2  # We created 2 shards

    @pytest.mark.asyncio
    async def test_stream_description(self, kinesis_config):
        """Test getting stream description."""
        from truthound.realtime.adapters.kinesis import KinesisAdapter

        adapter = KinesisAdapter(kinesis_config)

        async with adapter:
            description = await adapter.get_stream_description()
            assert description["StreamName"] == "test-stream"
            assert description["StreamStatus"] in ["ACTIVE", "CREATING"]

    @pytest.mark.asyncio
    async def test_reconnection(self, kinesis_config):
        """Test adapter reconnection."""
        from truthound.realtime.adapters.kinesis import KinesisAdapter

        adapter = KinesisAdapter(kinesis_config)

        await adapter.connect()
        assert adapter.is_connected

        # Simulate reconnection
        await adapter.reconnect()
        assert adapter.is_connected

        await adapter.disconnect()

    @pytest.mark.asyncio
    async def test_error_handling(self, kinesis_config):
        """Test error handling for invalid operations."""
        from truthound.realtime.adapters.kinesis import KinesisAdapter
        from truthound.realtime.protocols import StreamMessage, ConnectionError

        adapter = KinesisAdapter(kinesis_config)

        # Try to produce without connecting
        message = StreamMessage(
            key="test-pk",
            value={"test": True},
            partition=-1,
            offset=-1,
            timestamp=datetime.now(timezone.utc),
            headers=(),
            topic="test-stream",
        )

        with pytest.raises(ConnectionError):
            await adapter.produce(message)


@pytest.mark.skipif(
    not TESTCONTAINERS_AVAILABLE,
    reason="testcontainers not installed"
)
@pytest.mark.skipif(
    not AIOBOTOCORE_AVAILABLE,
    reason="aiobotocore not installed"
)
class TestKinesisStreamingPipeline:
    """Integration tests for complete streaming pipeline with Kinesis."""

    @pytest.fixture(scope="class")
    async def kinesis_env(self):
        """Start Kinesis environment for pipeline tests."""
        from truthound.realtime.testing.containers import StreamTestEnvironment

        async with StreamTestEnvironment(["kinesis"]) as env:
            await env.create_kinesis_stream("pipeline-stream", shard_count=2)
            yield env

    @pytest.mark.asyncio
    async def test_validation_pipeline(self, kinesis_env):
        """Test complete validation pipeline with Kinesis."""
        from truthound.realtime.adapters.kinesis import KinesisAdapter, KinesisAdapterConfig
        from truthound.realtime.protocols import StreamMessage
        from truthound.realtime.testing.fixtures import TestFixtures

        config = KinesisAdapterConfig(
            stream_name="pipeline-stream",
            region_name="us-east-1",
            endpoint_url=kinesis_env.get_service_url("kinesis"),
            aws_access_key_id="test",
            aws_secret_access_key="test",
        )

        adapter = KinesisAdapter(config)
        fixtures = TestFixtures()

        async with adapter:
            # Produce test messages
            test_data = fixtures.default_messages(20)
            for i, data in enumerate(test_data):
                message = StreamMessage(
                    key=f"pk-{i % 5}",
                    value=data,
                    partition=-1,
                    offset=-1,
                    timestamp=datetime.now(timezone.utc),
                    headers=(),
                    topic="pipeline-stream",
                )
                await adapter.produce(message)

            await adapter.flush()

            # Wait for messages
            await asyncio.sleep(3)

            # Consume and validate
            consumed = 0
            timeout = asyncio.get_event_loop().time() + 30

            async for msg in adapter.consume():
                consumed += 1
                value = msg.value

                # Basic validation check
                assert "id" in value or "value" in value

                if consumed >= 20 or asyncio.get_event_loop().time() > timeout:
                    break

            assert consumed >= 1  # At least some messages processed


@pytest.mark.skipif(
    not TESTCONTAINERS_AVAILABLE,
    reason="testcontainers not installed"
)
class TestLocalStackContainer:
    """Tests for LocalStack container management."""

    @pytest.mark.asyncio
    async def test_container_lifecycle(self):
        """Test LocalStack container start/stop."""
        from truthound.realtime.testing.containers import LocalStackTestContainer

        container = LocalStackTestContainer()

        # Start
        await container.start()
        assert container.is_running

        # Get URL
        url = container.get_connection_url()
        assert "http://" in url

        # Stop
        await container.stop()
        assert not container.is_running

    @pytest.mark.asyncio
    async def test_create_multiple_streams(self):
        """Test creating multiple Kinesis streams."""
        from truthound.realtime.testing.containers import LocalStackTestContainer

        async with LocalStackTestContainer() as container:
            # Create multiple streams
            await container.create_kinesis_stream("stream-1", shard_count=1)
            await container.create_kinesis_stream("stream-2", shard_count=2)
            await container.create_kinesis_stream("stream-3", shard_count=3)

            # All streams should be created successfully
            # (no exceptions thrown)



"""Tiered storage store implementation.

This module provides the main TieredStore class that manages multiple
storage tiers with automatic data migration.
"""

from __future__ import annotations

import json
import logging
from datetime import datetime
from typing import Any, Generic, TypeVar

from truthound.stores.base import (
    BaseStore,
    StoreConfig,
    StoreNotFoundError,
    StoreQuery,
    StoreReadError,
    StoreWriteError,
    ValidationStore,
)
from truthound.stores.results import ValidationResult
from truthound.stores.tiering.base import (
    InMemoryTierMetadataStore,
    StorageTier,
    TierInfo,
    TieringConfig,
    TieringResult,
    TierMetadataStore,
    TierMigrationError,
    TierNotFoundError,
)

ConfigT = TypeVar("ConfigT", bound=StoreConfig)

logger = logging.getLogger(__name__)


class TieredStore(ValidationStore[ConfigT], Generic[ConfigT]):
    """A store that manages multiple storage tiers.

    TieredStore provides automatic data placement and migration across
    multiple storage backends based on configurable policies.

    Example:
        >>> from truthound.stores import get_store
        >>> from truthound.stores.tiering import (
        ...     TieredStore,
        ...     StorageTier,
        ...     TieringConfig,
        ...     TierType,
        ...     AgeBasedTierPolicy,
        ... )
        >>>
        >>> # Define tiers
        >>> tiers = [
        ...     StorageTier(
        ...         name="hot",
        ...         store=get_store("filesystem", base_path=".truthound/hot"),
        ...         tier_type=TierType.HOT,
        ...         priority=1,
        ...     ),
        ...     StorageTier(
        ...         name="cold",
        ...         store=get_store("s3", bucket="archive-bucket"),
        ...         tier_type=TierType.COLD,
        ...         priority=2,
        ...     ),
        ... ]
        >>>
        >>> # Configure tiering
        >>> config = TieringConfig(
        ...     policies=[
        ...         AgeBasedTierPolicy("hot", "cold", after_days=30),
        ...     ],
        ... )
        >>>
        >>> # Create tiered store
        >>> store = TieredStore(tiers, config)
        >>>
        >>> # Save (goes to default tier)
        >>> store.save(result)
        >>>
        >>> # Get (searches all tiers)
        >>> result = store.get(run_id)
    """

    def __init__(
        self,
        tiers: list[StorageTier],
        config: TieringConfig | None = None,
        metadata_store: TierMetadataStore | None = None,
    ) -> None:
        """Initialize the tiered store.

        Args:
            tiers: List of storage tiers (must have at least one).
            config: Tiering configuration.
            metadata_store: Store for tier placement metadata.
        """
        if not tiers:
            raise ValueError("At least one tier must be provided")

        self._tiers = {t.name: t for t in tiers}
        self._tier_list = sorted(tiers, key=lambda t: t.priority)
        self._config = config or TieringConfig()
        self._metadata_store = metadata_store or InMemoryTierMetadataStore()
        self._initialized = False

        # Validate default tier exists
        if self._config.default_tier not in self._tiers:
            raise ValueError(
                f"Default tier '{self._config.default_tier}' not in tiers: "
                f"{list(self._tiers.keys())}"
            )

    @classmethod
    def _default_config(cls) -> ConfigT:
        """Create default configuration."""
        return StoreConfig()  # type: ignore

    def _do_initialize(self) -> None:
        """Initialize all tier stores."""
        for tier in self._tier_list:
            tier.store.initialize()

    def _get_tier(self, name: str) -> StorageTier:
        """Get a tier by name."""
        if name not in self._tiers:
            raise TierNotFoundError(name)
        return self._tiers[name]

    def _find_item_tier(self, item_id: str) -> StorageTier | None:
        """Find which tier contains an item.

        Searches tiers in priority order (fastest first).
        """
        # First check metadata
        info = self._metadata_store.get_info(item_id)
        if info and info.tier_name in self._tiers:
            tier = self._tiers[info.tier_name]
            if tier.store.exists(item_id):
                return tier

        # Fall back to searching all tiers
        for tier in self._tier_list:
            if tier.store.exists(item_id):
                return tier

        return None

    def _record_access(self, item_id: str) -> None:
        """Record an access for promotion consideration."""
        self._metadata_store.update_access(item_id)

        # Check if item should be promoted
        if self._config.enable_promotion:
            info = self._metadata_store.get_info(item_id)
            if info and info.access_count >= self._config.promotion_threshold:
                # Find higher-priority tier
                current_tier = self._tiers.get(info.tier_name)
                if current_tier:
                    for tier in self._tier_list:
                        if tier.priority < current_tier.priority:
                            # Promote to this tier
                            try:
                                self._migrate_item(
                                    item_id, info.tier_name, tier.name
                                )
                            except TierMigrationError as e:
                                logger.warning(f"Promotion failed: {e}")
                            break

    # -------------------------------------------------------------------------
    # CRUD Operations
    # -------------------------------------------------------------------------

    def save(
        self,
        item: ValidationResult,
        tier: str | None = None,
    ) -> str:
        """Save a validation result to a tier.

        Args:
            item: The validation result to save.
            tier: Target tier name (uses default if None).

        Returns:
            The run ID of the saved result.

        Raises:
            TierNotFoundError: If specified tier doesn't exist.
            StoreWriteError: If saving fails.
        """
        self.initialize()

        tier_name = tier or self._config.default_tier
        target_tier = self._get_tier(tier_name)

        # Save to tier
        item_id = target_tier.store.save(item)

        # Save metadata
        info = TierInfo(
            item_id=item_id,
            tier_name=tier_name,
            created_at=datetime.now(),
            size_bytes=len(json.dumps(item.to_dict(), default=str).encode()),
        )
        self._metadata_store.save_info(info)

        return item_id

    def get(self, item_id: str) -> ValidationResult:
        """Retrieve a validation result from any tier.

        Args:
            item_id: The run ID of the result.

        Returns:
            The validation result.

        Raises:
            StoreNotFoundError: If the result doesn't exist in any tier.
        """
        self.initialize()

        tier = self._find_item_tier(item_id)
        if not tier:
            raise StoreNotFoundError("ValidationResult", item_id)

        result = tier.store.get(item_id)

        # Record access for promotion
        self._record_access(item_id)

        return result

    def exists(self, item_id: str) -> bool:
        """Check if a validation result exists in any tier.

        Args:
            item_id: The run ID to check.

        Returns:
            True if the result exists.
        """
        self.initialize()
        return self._find_item_tier(item_id) is not None

    def delete(self, item_id: str) -> bool:
        """Delete a validation result from all tiers.

        Args:
            item_id: The run ID of the result to delete.

        Returns:
            True if something was deleted.
        """
        self.initialize()

        deleted = False

        # Delete from all tiers
        for tier in self._tier_list:
            if tier.store.delete(item_id):
                deleted = True

        # Delete metadata
        self._metadata_store.delete_info(item_id)

        return deleted

    def list_ids(self, query: StoreQuery | None = None) -> list[str]:
        """List validation result IDs matching the query.

        Combines results from all tiers, removing duplicates.
        """
        self.initialize()

        all_ids = set()
        for tier in self._tier_list:
            tier_ids = tier.store.list_ids(query)
            all_ids.update(tier_ids)

        return list(all_ids)

    def query(self, query: StoreQuery) -> list[ValidationResult]:
        """Query validation results from all tiers."""
        ids = self.list_ids(query)
        results = []

        for item_id in ids:
            try:
                result = self.get(item_id)
                results.append(result)
            except (StoreNotFoundError, StoreReadError):
                continue

        return results

    # -------------------------------------------------------------------------
    # Tier Operations
    # -------------------------------------------------------------------------

    def get_item_tier(self, item_id: str) -> str | None:
        """Get the current tier for an item.

        Args:
            item_id: The item identifier.

        Returns:
            Tier name, or None if not found.
        """
        self.initialize()

        info = self._metadata_store.get_info(item_id)
        if info:
            return info.tier_name

        tier = self._find_item_tier(item_id)
        return tier.name if tier else None

    def get_item_info(self, item_id: str) -> TierInfo | None:
        """Get tier information for an item.

        Args:
            item_id: The item identifier.

        Returns:
            Tier info, or None if not found.
        """
        self.initialize()
        return self._metadata_store.get_info(item_id)

    def move_to_tier(self, item_id: str, target_tier: str) -> bool:
        """Manually move an item to a specific tier.

        Args:
            item_id: The item identifier.
            target_tier: Name of the target tier.

        Returns:
            True if moved successfully.

        Raises:
            StoreNotFoundError: If item doesn't exist.
            TierNotFoundError: If target tier doesn't exist.
            TierMigrationError: If migration fails.
        """
        self.initialize()

        current_tier = self._find_item_tier(item_id)
        if not current_tier:
            raise StoreNotFoundError("ValidationResult", item_id)

        if current_tier.name == target_tier:
            return True  # Already in target tier

        self._get_tier(target_tier)  # Validate target exists

        self._migrate_item(item_id, current_tier.name, target_tier)
        return True

    def _migrate_item(
        self,
        item_id: str,
        from_tier: str,
        to_tier: str,
    ) -> None:
        """Migrate an item between tiers.

        Args:
            item_id: The item identifier.
            from_tier: Source tier name.
            to_tier: Destination tier name.

        Raises:
            TierMigrationError: If migration fails.
        """
        source = self._get_tier(from_tier)
        target = self._get_tier(to_tier)

        try:
            # Read from source
            item = source.store.get(item_id)

            # Write to target
            target.store.save(item)

            # Delete from source
            source.store.delete(item_id)

            # Update metadata
            info = self._metadata_store.get_info(item_id)
            if info:
                info.tier_name = to_tier
                info.migrated_at = datetime.now()
                self._metadata_store.save_info(info)
            else:
                new_info = TierInfo(
                    item_id=item_id,
                    tier_name=to_tier,
                    created_at=datetime.now(),
                    migrated_at=datetime.now(),
                )
                self._metadata_store.save_info(new_info)

            logger.info(f"Migrated {item_id} from {from_tier} to {to_tier}")

        except Exception as e:
            raise TierMigrationError(item_id, from_tier, to_tier, str(e))

    def run_tiering(self, dry_run: bool = False) -> TieringResult:
        """Run tier migration based on policies.

        Args:
            dry_run: If True, only report what would be migrated.

        Returns:
            Result of the tiering operation.
        """
        self.initialize()

        start_time = datetime.now()
        result = TieringResult(start_time=start_time, dry_run=dry_run)

        try:
            # Process each policy
            for policy in self._config.policies:
                from_tier = self._get_tier(policy.from_tier)

                # Get all items in source tier
                tier_items = self._metadata_store.list_by_tier(from_tier.name)
                result.items_scanned += len(tier_items)

                # Prepare batch context if needed
                if hasattr(policy, "prepare_batch"):
                    policy.prepare_batch(tier_items)

                # Evaluate each item
                for info in tier_items:
                    if policy.should_migrate(info):
                        migration = {
                            "item_id": info.item_id,
                            "from_tier": policy.from_tier,
                            "to_tier": policy.to_tier,
                            "size_bytes": info.size_bytes,
                        }

                        if dry_run:
                            result.migrations.append(migration)
                            result.items_migrated += 1
                            result.bytes_migrated += info.size_bytes
                        else:
                            try:
                                self._migrate_item(
                                    info.item_id,
                                    policy.from_tier,
                                    policy.to_tier,
                                )
                                result.migrations.append(migration)
                                result.items_migrated += 1
                                result.bytes_migrated += info.size_bytes
                            except TierMigrationError as e:
                                result.errors.append(str(e))

        except Exception as e:
            result.errors.append(f"Tiering failed: {str(e)}")
            logger.error(f"Tiering operation failed: {e}")

        result.end_time = datetime.now()
        return result

    def get_tier_stats(self) -> dict[str, Any]:
        """Get statistics for all tiers.

        Returns:
            Dictionary with tier statistics.
        """
        self.initialize()

        stats = {}
        for tier in self._tier_list:
            tier_items = self._metadata_store.list_by_tier(tier.name)
            total_size = sum(i.size_bytes for i in tier_items)
            total_accesses = sum(i.access_count for i in tier_items)

            stats[tier.name] = {
                "type": tier.tier_type.value,
                "priority": tier.priority,
                "item_count": len(tier_items),
                "total_size_bytes": total_size,
                "total_accesses": total_accesses,
                "cost_per_gb": tier.cost_per_gb,
                "estimated_monthly_cost": (
                    total_size / (1024**3) * tier.cost_per_gb
                ),
            }

        return stats

    def list_tiers(self) -> list[dict[str, Any]]:
        """List all configured tiers.

        Returns:
            List of tier information dictionaries.
        """
        return [tier.to_dict() for tier in self._tier_list]

    def close(self) -> None:
        """Close all tier stores."""
        for tier in self._tier_list:
            tier.store.close()

    @property
    def default_tier(self) -> str:
        """Get the default tier name."""
        return self._config.default_tier

    @property
    def tiers(self) -> dict[str, StorageTier]:
        """Get all tiers."""
        return self._tiers.copy()



"""Retention policy implementations.

This module provides various retention policies for different use cases.
"""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any, Callable

from truthound.stores.retention.base import (
    ItemMetadata,
    PolicyMode,
    RetentionAction,
    RetentionPolicy,
)


class TimeBasedPolicy(RetentionPolicy):
    """Retain items based on age.

    Items older than max_age will be marked for deletion.

    Example:
        >>> policy = TimeBasedPolicy(max_age_days=30)  # Keep 30 days
        >>> policy = TimeBasedPolicy(max_age_hours=72)  # Keep 72 hours
    """

    def __init__(
        self,
        max_age_days: int = 0,
        max_age_hours: int = 0,
        max_age_minutes: int = 0,
        action: RetentionAction = RetentionAction.DELETE,
    ) -> None:
        """Initialize the policy.

        Args:
            max_age_days: Maximum age in days.
            max_age_hours: Additional hours.
            max_age_minutes: Additional minutes.
            action: Action to take on expired items.
        """
        self.max_age = timedelta(
            days=max_age_days,
            hours=max_age_hours,
            minutes=max_age_minutes,
        )
        self._action = action

        if self.max_age.total_seconds() <= 0:
            raise ValueError("max_age must be positive")

    def should_retain(self, item: ItemMetadata) -> bool:
        """Check if item is younger than max_age."""
        age = datetime.now() - item.created_at
        return age <= self.max_age

    @property
    def description(self) -> str:
        """Human-readable description."""
        days = self.max_age.days
        hours = self.max_age.seconds // 3600
        minutes = (self.max_age.seconds % 3600) // 60

        parts = []
        if days:
            parts.append(f"{days} day{'s' if days != 1 else ''}")
        if hours:
            parts.append(f"{hours} hour{'s' if hours != 1 else ''}")
        if minutes:
            parts.append(f"{minutes} minute{'s' if minutes != 1 else ''}")

        return f"Keep items younger than {', '.join(parts)}"

    @property
    def action(self) -> RetentionAction:
        """Action to take on expired items."""
        return self._action

    def get_expiry_time(self, item: ItemMetadata) -> datetime:
        """Get when the item will expire."""
        return item.created_at + self.max_age

    def get_priority(self, item: ItemMetadata) -> int:
        """Older items have higher deletion priority."""
        age = datetime.now() - item.created_at
        return int(age.total_seconds())

    def to_dict(self) -> dict[str, Any]:
        """Serialize to dictionary."""
        result = super().to_dict()
        result.update(
            {
                "max_age_seconds": self.max_age.total_seconds(),
            }
        )
        return result


class CountBasedPolicy(RetentionPolicy):
    """Retain only a maximum number of items.

    When the count exceeds max_count, oldest items are marked for deletion.

    Example:
        >>> policy = CountBasedPolicy(max_count=1000)  # Keep max 1000
        >>> policy = CountBasedPolicy(max_count=100, per_asset=True)  # Per data asset
    """

    def __init__(
        self,
        max_count: int,
        per_asset: bool = False,
        action: RetentionAction = RetentionAction.DELETE,
    ) -> None:
        """Initialize the policy.

        Args:
            max_count: Maximum number of items to keep.
            per_asset: Apply limit per data_asset (vs globally).
            action: Action to take on excess items.
        """
        if max_count <= 0:
            raise ValueError("max_count must be positive")

        self.max_count = max_count
        self.per_asset = per_asset
        self._action = action

        # These are set during batch evaluation
        self._current_counts: dict[str, int] = {}
        self._item_ranks: dict[str, int] = {}

    def should_retain(self, item: ItemMetadata) -> bool:
        """Check if item should be retained based on count.

        Note: This requires batch context to work properly.
        Use RetentionStore.run_cleanup() which provides proper context.
        """
        # Check if we have rank information
        if item.item_id in self._item_ranks:
            rank = self._item_ranks[item.item_id]
            return rank <= self.max_count

        # Without batch context, we can't determine count-based retention
        return True

    def prepare_batch(self, items: list[ItemMetadata]) -> None:
        """Prepare for batch evaluation.

        Args:
            items: All items to consider for retention.
        """
        # Reset state
        self._current_counts = {}
        self._item_ranks = {}

        # Group by asset if needed
        if self.per_asset:
            asset_items: dict[str, list[ItemMetadata]] = {}
            for item in items:
                if item.data_asset not in asset_items:
                    asset_items[item.data_asset] = []
                asset_items[item.data_asset].append(item)

            for asset, asset_items_list in asset_items.items():
                # Sort by creation time (newest first)
                sorted_items = sorted(
                    asset_items_list,
                    key=lambda x: x.created_at,
                    reverse=True,
                )
                for rank, item in enumerate(sorted_items, 1):
                    self._item_ranks[item.item_id] = rank
                self._current_counts[asset] = len(asset_items_list)
        else:
            # Global count
            sorted_items = sorted(items, key=lambda x: x.created_at, reverse=True)
            for rank, item in enumerate(sorted_items, 1):
                self._item_ranks[item.item_id] = rank
            self._current_counts["__global__"] = len(items)

    @property
    def description(self) -> str:
        """Human-readable description."""
        scope = "per data asset" if self.per_asset else "globally"
        return f"Keep at most {self.max_count} items {scope}"

    @property
    def action(self) -> RetentionAction:
        """Action to take on excess items."""
        return self._action

    def get_priority(self, item: ItemMetadata) -> int:
        """Lower ranks (older) have higher priority for deletion."""
        rank = self._item_ranks.get(item.item_id, 0)
        if rank > self.max_count:
            return rank
        return 0

    def to_dict(self) -> dict[str, Any]:
        """Serialize to dictionary."""
        result = super().to_dict()
        result.update(
            {
                "max_count": self.max_count,
                "per_asset": self.per_asset,
            }
        )
        return result


class SizeBasedPolicy(RetentionPolicy):
    """Retain items until total size exceeds limit.

    When total size exceeds max_size, oldest items are marked for deletion.

    Example:
        >>> policy = SizeBasedPolicy(max_size_mb=500)  # Max 500 MB
        >>> policy = SizeBasedPolicy(max_size_gb=10, per_asset=True)  # 10 GB per asset
    """

    def __init__(
        self,
        max_size_bytes: int = 0,
        max_size_kb: int = 0,
        max_size_mb: int = 0,
        max_size_gb: int = 0,
        per_asset: bool = False,
        action: RetentionAction = RetentionAction.DELETE,
    ) -> None:
        """Initialize the policy.

        Args:
            max_size_bytes: Maximum size in bytes.
            max_size_kb: Maximum size in kilobytes.
            max_size_mb: Maximum size in megabytes.
            max_size_gb: Maximum size in gigabytes.
            per_asset: Apply limit per data_asset (vs globally).
            action: Action to take on excess items.
        """
        self.max_size = (
            max_size_bytes
            + max_size_kb * 1024
            + max_size_mb * 1024 * 1024
            + max_size_gb * 1024 * 1024 * 1024
        )
        self.per_asset = per_asset
        self._action = action

        if self.max_size <= 0:
            raise ValueError("max_size must be positive")

        # Batch state
        self._cumulative_sizes: dict[str, int] = {}
        self._item_cumulative: dict[str, int] = {}

    def should_retain(self, item: ItemMetadata) -> bool:
        """Check if item should be retained based on size."""
        if item.item_id in self._item_cumulative:
            cumulative = self._item_cumulative[item.item_id]
            return cumulative <= self.max_size

        return True

    def prepare_batch(self, items: list[ItemMetadata]) -> None:
        """Prepare for batch evaluation.

        Args:
            items: All items to consider for retention.
        """
        self._cumulative_sizes = {}
        self._item_cumulative = {}

        if self.per_asset:
            asset_items: dict[str, list[ItemMetadata]] = {}
            for item in items:
                if item.data_asset not in asset_items:
                    asset_items[item.data_asset] = []
                asset_items[item.data_asset].append(item)

            for asset, asset_items_list in asset_items.items():
                # Sort by creation time (newest first)
                sorted_items = sorted(
                    asset_items_list,
                    key=lambda x: x.created_at,
                    reverse=True,
                )
                cumulative = 0
                for item in sorted_items:
                    cumulative += item.size_bytes
                    self._item_cumulative[item.item_id] = cumulative
                self._cumulative_sizes[asset] = cumulative
        else:
            sorted_items = sorted(items, key=lambda x: x.created_at, reverse=True)
            cumulative = 0
            for item in sorted_items:
                cumulative += item.size_bytes
                self._item_cumulative[item.item_id] = cumulative
            self._cumulative_sizes["__global__"] = cumulative

    @property
    def description(self) -> str:
        """Human-readable description."""
        # Format size nicely
        if self.max_size >= 1024 * 1024 * 1024:
            size_str = f"{self.max_size / (1024 * 1024 * 1024):.1f} GB"
        elif self.max_size >= 1024 * 1024:
            size_str = f"{self.max_size / (1024 * 1024):.1f} MB"
        elif self.max_size >= 1024:
            size_str = f"{self.max_size / 1024:.1f} KB"
        else:
            size_str = f"{self.max_size} bytes"

        scope = "per data asset" if self.per_asset else "total"
        return f"Keep items within {size_str} {scope}"

    @property
    def action(self) -> RetentionAction:
        """Action to take on excess items."""
        return self._action

    def get_priority(self, item: ItemMetadata) -> int:
        """Items that push over limit have higher priority."""
        cumulative = self._item_cumulative.get(item.item_id, 0)
        if cumulative > self.max_size:
            return cumulative - self.max_size
        return 0

    def to_dict(self) -> dict[str, Any]:
        """Serialize to dictionary."""
        result = super().to_dict()
        result.update(
            {
                "max_size_bytes": self.max_size,
                "per_asset": self.per_asset,
            }
        )
        return result


class StatusBasedPolicy(RetentionPolicy):
    """Retain items based on validation status.

    Example:
        >>> # Delete failed results after 7 days, keep successes 30 days
        >>> policy = StatusBasedPolicy(
        ...     status="failure",
        ...     max_age_days=7,
        ... )
    """

    def __init__(
        self,
        status: str,
        max_age_days: int | None = None,
        retain: bool = True,
        action: RetentionAction = RetentionAction.DELETE,
    ) -> None:
        """Initialize the policy.

        Args:
            status: Status to match.
            max_age_days: Optional max age for matching items.
            retain: If True, keep matching items. If False, delete them.
            action: Action to take on non-retained items.
        """
        self.status = status
        self.max_age_days = max_age_days
        self.retain = retain
        self._action = action

    def should_retain(self, item: ItemMetadata) -> bool:
        """Check if item should be retained based on status."""
        matches_status = item.status == self.status

        if not matches_status:
            # Different status, use default retention
            return True

        # Status matches
        if self.max_age_days is not None:
            age = datetime.now() - item.created_at
            if age.days > self.max_age_days:
                return not self.retain

        return self.retain

    @property
    def description(self) -> str:
        """Human-readable description."""
        action = "Keep" if self.retain else "Delete"
        age_str = ""
        if self.max_age_days is not None:
            age_str = f" older than {self.max_age_days} days"
        return f"{action} items with status '{self.status}'{age_str}"

    @property
    def action(self) -> RetentionAction:
        """Action to take on non-retained items."""
        return self._action

    def to_dict(self) -> dict[str, Any]:
        """Serialize to dictionary."""
        result = super().to_dict()
        result.update(
            {
                "status": self.status,
                "max_age_days": self.max_age_days,
                "retain": self.retain,
            }
        )
        return result


class TagBasedPolicy(RetentionPolicy):
    """Retain items based on tags.

    Example:
        >>> # Keep items tagged as 'production'
        >>> policy = TagBasedPolicy(required_tags={"env": "production"})
        >>>
        >>> # Delete items tagged as 'temp'
        >>> policy = TagBasedPolicy(delete_tags={"type": "temp"})
    """

    def __init__(
        self,
        required_tags: dict[str, str] | None = None,
        delete_tags: dict[str, str] | None = None,
        any_match: bool = False,
        action: RetentionAction = RetentionAction.DELETE,
    ) -> None:
        """Initialize the policy.

        Args:
            required_tags: Tags that must be present to retain.
            delete_tags: Tags that trigger deletion if present.
            any_match: If True, match any tag. If False, match all.
            action: Action to take on non-retained items.
        """
        self.required_tags = required_tags or {}
        self.delete_tags = delete_tags or {}
        self.any_match = any_match
        self._action = action

    def should_retain(self, item: ItemMetadata) -> bool:
        """Check if item should be retained based on tags."""
        # Check delete tags first
        if self.delete_tags:
            matches = []
            for key, value in self.delete_tags.items():
                matches.append(item.tags.get(key) == value)

            if self.any_match and any(matches):
                return False
            if not self.any_match and all(matches):
                return False

        # Check required tags
        if self.required_tags:
            matches = []
            for key, value in self.required_tags.items():
                matches.append(item.tags.get(key) == value)

            if self.any_match:
                return any(matches)
            return all(matches)

        return True

    @property
    def description(self) -> str:
        """Human-readable description."""
        parts = []
        if self.required_tags:
            tags_str = ", ".join(f"{k}={v}" for k, v in self.required_tags.items())
            match_str = "any of" if self.any_match else "all of"
            parts.append(f"Keep items with {match_str} tags: {tags_str}")
        if self.delete_tags:
            tags_str = ", ".join(f"{k}={v}" for k, v in self.delete_tags.items())
            match_str = "any of" if self.any_match else "all of"
            parts.append(f"Delete items with {match_str} tags: {tags_str}")
        return "; ".join(parts) or "No tag rules"

    @property
    def action(self) -> RetentionAction:
        """Action to take on non-retained items."""
        return self._action

    def to_dict(self) -> dict[str, Any]:
        """Serialize to dictionary."""
        result = super().to_dict()
        result.update(
            {
                "required_tags": self.required_tags,
                "delete_tags": self.delete_tags,
                "any_match": self.any_match,
            }
        )
        return result


class CompositePolicy(RetentionPolicy):
    """Combine multiple policies with custom logic.

    Example:
        >>> # Keep if (age < 30 days) AND (count < 1000)
        >>> policy = CompositePolicy(
        ...     policies=[
        ...         TimeBasedPolicy(max_age_days=30),
        ...         CountBasedPolicy(max_count=1000),
        ...     ],
        ...     mode=PolicyMode.ALL,
        ... )
    """

    def __init__(
        self,
        policies: list[RetentionPolicy],
        mode: PolicyMode = PolicyMode.ALL,
        action: RetentionAction = RetentionAction.DELETE,
    ) -> None:
        """Initialize the policy.

        Args:
            policies: Child policies to combine.
            mode: How to combine results (ALL = all must agree, ANY = any can keep).
            action: Action to take on non-retained items.
        """
        self.policies = policies
        self.mode = mode
        self._action = action

    def should_retain(self, item: ItemMetadata) -> bool:
        """Check if item should be retained based on all child policies."""
        results = [p.should_retain(item) for p in self.policies]

        if self.mode == PolicyMode.ALL:
            return all(results)
        return any(results)

    def prepare_batch(self, items: list[ItemMetadata]) -> None:
        """Prepare child policies for batch evaluation."""
        for policy in self.policies:
            if hasattr(policy, "prepare_batch"):
                policy.prepare_batch(items)

    @property
    def description(self) -> str:
        """Human-readable description."""
        mode_str = "all of" if self.mode == PolicyMode.ALL else "any of"
        child_descs = [f"  - {p.description}" for p in self.policies]
        return f"Retain if {mode_str}:\n" + "\n".join(child_descs)

    @property
    def action(self) -> RetentionAction:
        """Action to take on non-retained items."""
        return self._action

    def get_priority(self, item: ItemMetadata) -> int:
        """Get combined priority from all policies."""
        return sum(p.get_priority(item) for p in self.policies)

    def to_dict(self) -> dict[str, Any]:
        """Serialize to dictionary."""
        result = super().to_dict()
        result.update(
            {
                "policies": [p.to_dict() for p in self.policies],
                "mode": self.mode.value,
            }
        )
        return result


class CustomPolicy(RetentionPolicy):
    """Custom retention policy using a callable.

    Example:
        >>> def keep_weekday_only(item: ItemMetadata) -> bool:
        ...     return item.created_at.weekday() < 5  # Mon-Fri
        >>>
        >>> policy = CustomPolicy(
        ...     predicate=keep_weekday_only,
        ...     description="Keep items created on weekdays",
        ... )
    """

    def __init__(
        self,
        predicate: Callable[[ItemMetadata], bool],
        description: str = "Custom policy",
        action: RetentionAction = RetentionAction.DELETE,
    ) -> None:
        """Initialize the policy.

        Args:
            predicate: Function that returns True to retain, False to delete.
            description: Human-readable description.
            action: Action to take on non-retained items.
        """
        self._predicate = predicate
        self._description = description
        self._action = action

    def should_retain(self, item: ItemMetadata) -> bool:
        """Check if item should be retained using custom predicate."""
        return self._predicate(item)

    @property
    def description(self) -> str:
        """Human-readable description."""
        return self._description

    @property
    def action(self) -> RetentionAction:
        """Action to take on non-retained items."""
        return self._action



"""Tests for Elasticsearch data source."""

from __future__ import annotations

import pytest
from unittest.mock import AsyncMock, MagicMock, patch

from truthound.datasources._protocols import ColumnType, DataSourceCapability
from truthound.datasources.nosql.elasticsearch import (
    ElasticsearchDataSource,
    ElasticsearchConfig,
    ElasticsearchError,
    ElasticsearchConnectionError,
)


class TestElasticsearchConfig:
    """Tests for ElasticsearchConfig."""

    def test_default_values(self) -> None:
        """Test default configuration values."""
        config = ElasticsearchConfig(index="test-index")

        assert config.hosts == ["http://localhost:9200"]
        assert config.index == "test-index"
        assert config.cloud_id is None
        assert config.verify_certs is True
        assert config.scroll_timeout == "5m"
        assert config.scroll_size == 1000

    def test_custom_values(self) -> None:
        """Test custom configuration values."""
        config = ElasticsearchConfig(
            hosts=["http://node1:9200", "http://node2:9200"],
            index="my-index-*",
            scroll_size=500,
            request_timeout=60,
        )

        assert len(config.hosts) == 2
        assert config.index == "my-index-*"
        assert config.scroll_size == 500
        assert config.request_timeout == 60


class TestElasticsearchDataSource:
    """Tests for ElasticsearchDataSource."""

    def test_requires_index(self) -> None:
        """Test that index is required."""
        with pytest.raises(ElasticsearchError, match="Index name is required"):
            ElasticsearchDataSource(ElasticsearchConfig())

    def test_creation(self) -> None:
        """Test source creation."""
        config = ElasticsearchConfig(index="test-index")
        source = ElasticsearchDataSource(config)

        assert source.source_type == "elasticsearch"
        assert source.index == "test-index"

    def test_name_property(self) -> None:
        """Test name property."""
        config = ElasticsearchConfig(index="my-index")
        source = ElasticsearchDataSource(config)

        assert source.name == "es://my-index"

    def test_name_property_custom(self) -> None:
        """Test custom name."""
        config = ElasticsearchConfig(
            name="custom_name",
            index="idx",
        )
        source = ElasticsearchDataSource(config)

        assert source.name == "custom_name"

    def test_capabilities(self) -> None:
        """Test capabilities include expected values."""
        config = ElasticsearchConfig(index="idx")
        source = ElasticsearchDataSource(config)
        caps = source.capabilities

        assert DataSourceCapability.SCHEMA_INFERENCE in caps
        assert DataSourceCapability.SAMPLING in caps
        assert DataSourceCapability.STREAMING in caps
        assert DataSourceCapability.ROW_COUNT in caps

    def test_from_cloud(self) -> None:
        """Test factory method for Elastic Cloud."""
        source = ElasticsearchDataSource.from_cloud(
            cloud_id="my-deployment:base64string",
            api_key="my-api-key",
            index="logs-*",
        )

        assert source.index == "logs-*"
        assert source.config.cloud_id == "my-deployment:base64string"
        assert source.config.api_key == "my-api-key"

    def test_from_hosts(self) -> None:
        """Test factory method from host list."""
        source = ElasticsearchDataSource.from_hosts(
            hosts=["http://node1:9200", "http://node2:9200"],
            index="my-index",
            username="elastic",
            password="changeme",
        )

        assert source.index == "my-index"
        assert len(source.config.hosts) == 2
        assert source.config.username == "elastic"

    def test_es_type_mapping(self) -> None:
        """Test Elasticsearch type to ColumnType mapping."""
        mapping = ElasticsearchDataSource.ES_TYPE_MAPPING

        assert mapping["text"] == ColumnType.STRING
        assert mapping["keyword"] == ColumnType.STRING
        assert mapping["long"] == ColumnType.INTEGER
        assert mapping["integer"] == ColumnType.INTEGER
        assert mapping["double"] == ColumnType.FLOAT
        assert mapping["boolean"] == ColumnType.BOOLEAN
        assert mapping["date"] == ColumnType.DATETIME
        assert mapping["binary"] == ColumnType.BINARY
        assert mapping["object"] == ColumnType.STRUCT
        assert mapping["nested"] == ColumnType.STRUCT

    def test_parse_mapping_properties(self) -> None:
        """Test parsing Elasticsearch mappings."""
        config = ElasticsearchConfig(index="test")
        source = ElasticsearchDataSource(config)

        properties = {
            "title": {"type": "text"},
            "count": {"type": "integer"},
            "price": {"type": "double"},
            "active": {"type": "boolean"},
            "created": {"type": "date"},
        }

        schema: dict[str, ColumnType] = {}
        source._parse_mapping_properties(properties, "", schema)

        assert schema["title"] == ColumnType.STRING
        assert schema["count"] == ColumnType.INTEGER
        assert schema["price"] == ColumnType.FLOAT
        assert schema["active"] == ColumnType.BOOLEAN
        assert schema["created"] == ColumnType.DATETIME

    def test_parse_nested_mapping(self) -> None:
        """Test parsing nested Elasticsearch mappings."""
        config = ElasticsearchConfig(index="test", flatten_nested=True)
        source = ElasticsearchDataSource(config)

        properties = {
            "user": {
                "properties": {
                    "name": {"type": "keyword"},
                    "age": {"type": "integer"},
                }
            },
        }

        schema: dict[str, ColumnType] = {}
        source._parse_mapping_properties(properties, "", schema)

        assert "user.name" in schema
        assert "user.age" in schema
        assert schema["user.name"] == ColumnType.STRING
        assert schema["user.age"] == ColumnType.INTEGER

    def test_parse_nested_mapping_no_flatten(self) -> None:
        """Test parsing nested mappings without flattening."""
        config = ElasticsearchConfig(index="test", flatten_nested=False)
        source = ElasticsearchDataSource(config)

        properties = {
            "user": {
                "properties": {
                    "name": {"type": "keyword"},
                }
            },
        }

        schema: dict[str, ColumnType] = {}
        source._parse_mapping_properties(properties, "", schema)

        assert "user" in schema
        assert schema["user"] == ColumnType.STRUCT

    @pytest.mark.asyncio
    async def test_sample_async(self) -> None:
        """Test async sampling creates new config."""
        config = ElasticsearchConfig(
            index="test-index",
            max_documents=10000,
        )
        source = ElasticsearchDataSource(config)

        sampled = await source.sample_async(100)

        assert sampled.config.max_documents == 100
        assert "sample" in sampled.name


class TestElasticsearchExceptions:
    """Tests for Elasticsearch exceptions."""

    def test_elasticsearch_error(self) -> None:
        """Test ElasticsearchError."""
        error = ElasticsearchError("Test error")
        assert str(error) == "Test error"

    def test_elasticsearch_connection_error(self) -> None:
        """Test ElasticsearchConnectionError."""
        error = ElasticsearchConnectionError(
            "Connection refused",
            hosts=["http://localhost:9200"],
        )
        assert error.hosts == ["http://localhost:9200"]
        assert "Elasticsearch connection failed" in str(error)
        assert "Connection refused" in str(error)
